'''Inspiration from the deskbar pidgin plugin and from the gajim kupfer
plugin'''
__kupfer_name__ = _("Pidgin")
__kupfer_sources__ = ("ContactsSource", )
__kupfer_actions__ = (
    "OpenChat",
    "SendMessage",
)
__description__ = _("Access to Pidgin Contacts")
__author__ = ("Chmouel Boudjnah <chmouel@chmouel.com>, "
              "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>")
__version__ = "2017.1"

import dbus

from kupfer.objects import Action, TextLeaf, TextSource
from kupfer import pretty, scheduler
from kupfer import icons
from kupfer import plugin_support
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.weaklib import dbus_signal_connect_weakly
from kupfer.obj.contacts import NAME_KEY, EMAIL_KEY, ContactLeaf, is_valid_email

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "show_offline",
        "label": _("Show offline contacts"),
        "type": bool,
        "value": False,
    },
)

plugin_support.check_dbus_connection()

# Contact data contstants
PIDGIN_ACCOUNT = "_PIDGIN_ACCOUNT"
PIDGIN_JID = "_PIDGIN_JID"

# D-Bus "addresses"
SERVICE_NAME = "im.pidgin.purple.PurpleService"
OBJECT_NAME = "/im/pidgin/purple/PurpleObject"
IFACE_NAME = "im.pidgin.purple.PurpleInterface"

def _create_dbus_connection(activate=False):
    ''' Create dbus connection to Pidgin
    @activate: true=starts pidgin if not running
    '''
    interface = None
    obj = None
    sbus = dbus.SessionBus()

    try:
        #check for running pidgin (code from note.py)
        proxy_obj = sbus.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus')
        dbus_iface = dbus.Interface(proxy_obj, 'org.freedesktop.DBus')
        if activate or dbus_iface.NameHasOwner(SERVICE_NAME):
            obj = sbus.get_object(SERVICE_NAME, OBJECT_NAME)
        if obj:
            interface = dbus.Interface(obj, IFACE_NAME)
    except dbus.exceptions.DBusException as err:
        pretty.print_debug(err)
    return interface


def _send_message_to_contact(pcontact, message, present=False):
    """
    Send @message to PidginContact @pcontact
    """
    interface = _create_dbus_connection()
    if not interface:
        return
    account = pcontact[PIDGIN_ACCOUNT]
    jid = pcontact[PIDGIN_JID]
    conversation = interface.PurpleConversationNew(1, account, jid)
    im = interface.PurpleConvIm(conversation)
    interface.PurpleConvImSend(im, message)
    if present:
        interface.PurpleConversationPresent(conversation)

class ContactAction (Action):
    def get_required_slots(self):
        return ()
    def item_types(self):
        yield ContactLeaf
    def valid_for_item(self, leaf):
        return all(slot in leaf for slot in self.get_required_slots())

class OpenChat(ContactAction):
    """ Open Chat Conversation Window with jid """
    # consider it as main action for pidgin contacts
    rank_adjust = 5

    def __init__(self):
        Action.__init__(self, _('Open Chat'))

    def activate(self, leaf):
        _send_message_to_contact(leaf, "", present=True)

    def get_required_slots(self):
        return [PIDGIN_ACCOUNT, PIDGIN_JID]

class ChatTextSource (TextSource):
    def get_rank(self):
        return 100
    def get_text_items(self, text):
        n = len(text)
        summary = text[:10] + (text[10:11] and "..")
        desc_template = ngettext("%s (%d character)", "%s (%d characters)", n)
        yield TextLeaf(text, desc_template % (summary, n))

    def get_items(self, text):
        return self.get_text_items(text)

class SendMessage (ContactAction):
    """ Send chat message directly from Kupfer """
    def __init__(self):
        Action.__init__(self, _("Send Message..."))

    def activate(self, leaf, iobj):
        _send_message_to_contact(leaf, iobj.object)

    def get_required_slots(self):
        return [PIDGIN_ACCOUNT, PIDGIN_JID]

    def requires_object(self):
        return True
    def object_types(self):
        yield TextLeaf
    def object_source(self, for_item=None):
        return TextSource()
    def valid_object(self, iobj, for_item=None):
        # ugly, but we don't want derived text
        return type(iobj) is TextLeaf

class PidginContact(ContactLeaf):
    """ Leaf represent single contact from Pidgin """
    grouping_slots = ContactLeaf.grouping_slots + (EMAIL_KEY, )
    def __init__(self, jid, name, account, icon, protocol, available,
        status_message):
        slots = {
            EMAIL_KEY: jid,
            NAME_KEY: name or jid,
            PIDGIN_ACCOUNT: account,
            PIDGIN_JID: jid,
        }
        if not is_valid_email(jid):
            slots[EMAIL_KEY] = None

        ContactLeaf.__init__(self, slots, name or jid)

        self.kupfer_add_alias(jid)

        self._description = _("[%(status)s] %(userid)s/%(service)s") % \
                {
                    "status": _("Available") if available else _("Away"),
                    "userid": jid,
                    "service": protocol,
                }

        if status_message:
            self._description += "\n%s" % status_message

        self.account = account
        self.jid = jid
        self.icon = icon
        self.protocol = protocol

    def repr_key(self):
        # the repr key should be persistent and hopefully unique
        return "%s, %s" % (self.protocol, self.object[PIDGIN_JID])

    def get_description(self):
        return self._description

    def get_thumbnail(self, width, height):
        if not self.icon:
            return
        return icons.get_pixbuf_from_file(self.icon, width, height)

    def get_icon_name(self):
        return "stock_person"


class ContactsSource(AppLeafContentMixin, ToplevelGroupingSource):
    ''' Get contacts from all on-line accounts in Pidgin via DBus '''
    appleaf_content_id = 'pidgin'
    source_use_cache = False

    def __init__(self):
        ToplevelGroupingSource.__init__(self, _('Pidgin Contacts'), "Contacts")
        self._version = 5
        self.all_buddies = {}

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        self._install_dbus_signal()
        self._buddy_update_timer = scheduler.Timer()
        self._buddy_update_queue = set()

    def _get_pidgin_contact(self, interface, buddy, account=None, protocol=None):
        if not account:
            account = interface.PurpleBuddyGetAccount(buddy)

        if not protocol:
            protocol = interface.PurpleAccountGetProtocolName(account)

        jid = interface.PurpleBuddyGetName(buddy)
        name = interface.PurpleBuddyGetAlias(buddy)
        _icon = interface.PurpleBuddyGetIcon(buddy)
        icon = None
        if _icon != 0:
            icon = interface.PurpleBuddyIconGetFullPath(_icon)
        presenceid = interface.PurpleBuddyGetPresence(buddy)
        statusid = interface.PurplePresenceGetActiveStatus(presenceid)
        availability = interface.PurplePresenceIsAvailable(presenceid)
        status_message = interface.PurpleStatusGetAttrString(statusid, "message")

        return PidginContact(jid, name, account, icon, protocol, availability,
                status_message)

    def _get_all_buddies(self):
        interface = _create_dbus_connection()
        if interface is None:
            return

        accounts = interface.PurpleAccountsGetAllActive()
        show_offline = __kupfer_settings__["show_offline"]
        for account in accounts:
            buddies = interface.PurpleFindBuddies(account, dbus.String(''))
            protocol = interface.PurpleAccountGetProtocolName(account)

            for buddy in buddies:
                if not (show_offline or interface.PurpleBuddyIsOnline(buddy)):
                    continue

                self.all_buddies[buddy] = self._get_pidgin_contact(interface,
                        buddy, protocol=protocol, account=account)

    def _remove_buddies_not_connected(self):
        """ Remove buddies that belong to accounts no longer connected """
        if not self.all_buddies:
            return
        interface = _create_dbus_connection()
        if interface is None:
            self.all_buddies = {}
            return

        try:
            # extra careful as this will fail when Pidgin is Quitting
            accounts = interface.PurpleAccountsGetAllActive()
        except dbus.DBusException:
            self.all_buddies = {}
            return
        is_disconnected = interface.PurpleAccountIsDisconnected
        conn_accounts = set(a for a in accounts if not is_disconnected(a))
        for buddy, pcontact in list(self.all_buddies.items()):
            if pcontact.account not in conn_accounts:
                del self.all_buddies[buddy]

    def _signing_off(self, conn):
        self.output_debug("Pidgin Signing Off", conn)
        self._remove_buddies_not_connected()
        self.mark_for_update()

    def _update_pending(self):
        """Update all buddies in the update queue"""
        interface = _create_dbus_connection()
        if interface is None:
            self._buddy_update_queue.clear()
            return
        show_offline = __kupfer_settings__["show_offline"]
        for buddy in self._buddy_update_queue:
            if show_offline or interface.PurpleBuddyIsOnline(buddy):
                self.output_debug("updating buddy", buddy)
                pcontact = self._get_pidgin_contact(interface, buddy)
                self.all_buddies[buddy] = pcontact
            else:
                self.all_buddies.pop(buddy, None)
        self._buddy_update_queue.clear()
        self.mark_for_update()

    def _buddy_needs_update(self, buddy):
        """add @buddy to the update queue"""
        if self._buddy_update_queue is not None:
            self._buddy_update_queue.add(buddy)
            self._buddy_update_timer.set(1, self._update_pending)

    def _buddy_signed_on(self, buddy):
        if buddy not in self.all_buddies:
            self._buddy_needs_update(buddy)

    def _buddy_signed_off(self, buddy):
        if buddy in self.all_buddies:
            del self.all_buddies[buddy]
            self.mark_for_update()

    def _buddy_status_changed(self, buddy, old, new):
        '''Callback when status is changed reload the entry
        which get the new status'''
        self._buddy_needs_update(buddy)

    def _install_dbus_signal(self):
        '''Add signals to pidgin when buddy goes offline or
        online to update the list'''
        try:
            session_bus = dbus.Bus()
        except dbus.DBusException:
            return

        dbus_signal_connect_weakly(session_bus, "SigningOff",
                self._signing_off, dbus_interface=IFACE_NAME)

        dbus_signal_connect_weakly(session_bus, "BuddySignedOn",
                self._buddy_signed_on, dbus_interface=IFACE_NAME)

        dbus_signal_connect_weakly(session_bus, "BuddyStatusChanged",
                self._buddy_status_changed, dbus_interface=IFACE_NAME)

        dbus_signal_connect_weakly(session_bus, "BuddySignedOff",
                self._buddy_signed_off, dbus_interface=IFACE_NAME)

    def get_items(self):
        if not self.all_buddies:
            self._get_all_buddies()
        return list(self.all_buddies.values())

    def should_sort_lexically(self):
        return True

    def get_icon_name(self):
        return 'pidgin'

    def provides(self):
        yield PidginContact



"""
This module contains Helper constructs

This module is a part of the program Kupfer, see the main program file for
more information.
"""

from gi.repository import Gio, GLib

from kupfer import pretty

class PicklingHelperMixin (object):
    """ This pickling helper will define __getstate__/__setstate__
    acting simply on the class dictionary; it is up to the inheriting
    class to set up:
    pickle_prepare:
        Modify the instance dict to remove any unpickleable attributes,
        the resulting dict will be pickled
    unpickle_finish:
        Finish unpickling by restoring nonpickled attributes from the
        saved class dict, or setting up change callbacks or similar
    """
    def pickle_prepare(self):
        pass
    def unpickle_finish(self):
        pass
    def __getstate__(self):
        """On pickle, getstate will call self.pickle_prepare(),
        then it will return the class' current __dict__
        """
        self.pickle_prepare()
        return self.__dict__

    def __setstate__(self, state):
        """On unpickle, setstate will restore the class' __dict__,
        then call self.unpickle_finish()
        """
        self.__dict__.update(state)
        self.unpickle_finish()

class NonpersistentToken (PicklingHelperMixin):
    """A token will keep a reference until pickling, when it is deleted"""
    def __init__(self, data):
        self.data = data
    def __bool__(self):
        return self.data
    def pickle_prepare(self):
        self.data = None

class FilesystemWatchMixin (object):
    """A mixin for Sources watching directories"""

    def monitor_directories(self, *directories, **kwargs):
        """Register @directories for monitoring;

        On changes, the Source will be marked for update.
        This method returns a monitor token that has to be
        stored for the monitor to be active.

        The token will be a false value if nothing could be monitored.

        Nonexisting directories are skipped, if not passing
        the kwarg @force
        """
        tokens = []
        force = kwargs.get('force', False)
        for directory in directories:
            gfile = Gio.File.new_for_path(directory)
            if not force and not gfile.query_exists():
                continue
            try:
                monitor = gfile.monitor_directory(Gio.FileMonitorFlags.NONE, None)
            except GLib.GError as exc:
                pretty.print_debug(__name__, "FilesystemWatchMixin", exc)
                continue
            if monitor:
                monitor.connect("changed", self.__directory_changed)
                tokens.append(monitor)
        return NonpersistentToken(tokens)

    def monitor_include_file(self, gfile):
        """Return whether @gfile should trigger an update event
        by default, files beginning with "." are ignored
        """
        return not (gfile and gfile.get_basename().startswith("."))

    def __directory_changed(self, monitor, file1, file2, evt_type):
        if (evt_type in (Gio.FileMonitorEvent.CREATED,
                Gio.FileMonitorEvent.DELETED) and
                self.monitor_include_file(file1)):
            self.mark_for_update()

def reverse_action(action, rank=0):
    """Return a reversed version a three-part action

    @action: the action class
    @rank: the rank_adjust to give the reversed action

    A three-part action requires a direct object (item) and an indirect
    object (iobj).

    In general, the item must be from the Catalog, while the iobj can be
    from one, specified special Source. If this is used, and the action
    will be reversed, the base action must be the one specifying a
    source for the iobj. The reversed action will always take both item
    and iobj from the Catalog, filtered by type.

    If valid_object(iobj, for_leaf=None) is used, it will always be
    called with only the new item as the first parameter when reversed.
    """
    class ReverseAction (action):
        rank_adjust = rank
        def activate(self, leaf, iobj):
            return action.activate(self, iobj, leaf)
        def item_types(self):
            return action.object_types(self)
        def valid_for_item(self, leaf):
            try:
                return action.valid_object(self, leaf)
            except AttributeError:
                return True
        def object_types(self):
            return action.item_types(self)
        def valid_object(self, obj, for_item=None):
            return action.valid_for_item(self, obj)
        def object_source(self, for_item=None):
            return None
    ReverseAction.__name__ = "Reverse" + action.__name__
    return ReverseAction




__kupfer_name__ = _("Default Email Client")
__kupfer_actions__ = (
    "NewMailAction",
)
__description__ = _("Compose email using the system's default mailto: handler")
__version__ = "2017.1"
__author__ = ""

from kupfer.objects import Action
from kupfer.objects import TextLeaf, UrlLeaf
from kupfer.obj.contacts import ContactLeaf, email_from_leaf
from kupfer import utils


class NewMailAction(Action):
    def __init__(self):
        Action.__init__(self, _('Compose Email To'))

    def activate(self, leaf):
        email = email_from_leaf(leaf)
        utils.show_url("mailto:%s" % email)

    def activate_multiple(self, objects):
        recipients = ",".join(email_from_leaf(L) for L in objects)
        url = "mailto:" + recipients
        utils.show_url(url)

    def item_types(self):
        yield ContactLeaf
        yield TextLeaf
        yield UrlLeaf
    def valid_for_item(self, item):
        return bool(email_from_leaf(item))

    def get_description(self):
        return __description__
    def get_icon_name(self):
        return "mail-message-new"



# -*- coding: utf-8 -*

__kupfer_name__ = _("LXQT Session Management")
__kupfer_sources__ = ("LxqtItemsSource", )
__description__ = _("Actions for LXQT desktop")
__version__ = "2020-08-23"
__author__ = "Leonardo Masuero <leom255255@gmail.com>"
# Based on XFCE Session Management by Karol Będkowski

from kupfer.plugin import session_support as support


# sequences of argument lists
LOGOUT_CMD = (["lxqt-leave", "--logout"],)
SHUTDOWN_CMD = (["lxqt-leave", "--shutdown"],)
LOCKSCREEN_CMD = (["lxqt-leave", "--lockscreen"], )


class LxqtItemsSource (support.CommonSource):
    def __init__(self):
        support.CommonSource.__init__(self, _("LXQT Session Management"))
    def get_items(self):
        return (
            support.Logout(LOGOUT_CMD),
            support.LockScreen(LOCKSCREEN_CMD),
            support.Shutdown(SHUTDOWN_CMD),
        )



__kupfer_name__ = _("Document Templates")
__kupfer_sources__ = ("TemplatesSource", )
__kupfer_actions__ = ("CreateNewDocument", )
__description__ = _("Create new documents from your templates")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os

from gi.repository import Gio, GLib

from kupfer.objects import Leaf, Action, Source, FileLeaf
from kupfer import icons, utils, pretty
from kupfer.obj import helplib
from kupfer.obj.helplib import FilesystemWatchMixin

DEFAULT_TMPL_DIR = "~/Templates"

class Template (FileLeaf):
    def __init__(self, path):
        basename = GLib.filename_display_basename(path)
        nameroot, ext = os.path.splitext(basename)
        FileLeaf.__init__(self, path, _("%s template") % nameroot)

    def get_actions(self):
        yield CreateDocumentIn()
        for a in FileLeaf.get_actions(self):
            yield a

    def get_gicon(self):
        file_gicon = FileLeaf.get_gicon(self)
        return icons.ComposedIcon("text-x-generic-template", file_gicon)

class EmptyFile (Leaf):
    def __init__(self):
        Leaf.__init__(self, None, _("Empty File"))
    def repr_key(self):
        return ""
    def get_actions(self):
        yield CreateDocumentIn()
    def get_icon_name(self):
        return "text-x-generic"

class NewFolder (Leaf):
    def __init__(self):
        Leaf.__init__(self, None, _("New Folder"))
    def repr_key(self):
        return ""
    def get_actions(self):
        yield CreateDocumentIn()
    def get_icon_name(self):
        return "folder"

class CreateNewDocument (Action):
    def __init__(self):
        Action.__init__(self, _("Create New Document..."))

    def has_result(self):
        return True
    def activate(self, leaf, iobj):
        if iobj.object is not None:
            # Copy the template to destination directory
            basename = os.path.basename(iobj.object)
            tmpl_gfile = Gio.File.new_for_path(iobj.object)
            destpath = utils.get_destpath_in_directory(leaf.object, basename)
            destfile = Gio.File.new_for_path(destpath)
            tmpl_gfile.copy(destfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None)
        elif isinstance(iobj, NewFolder):
            filename = str(iobj)
            destpath = utils.get_destpath_in_directory(leaf.object, filename)
            os.makedirs(destpath)
        else:
            # create new empty file
            filename = str(iobj)
            f, destpath = utils.get_destfile_in_directory(leaf.object, filename)
            f.close()
        return FileLeaf(destpath)

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, leaf):
        return leaf.is_dir()

    def requires_object(self):
        return True
    def object_types(self):
        yield Template
        yield EmptyFile
        yield NewFolder

    def object_source(self, for_item=None):
        return TemplatesSource()

    def get_description(self):
        return _("Create a new document from template")
    def get_icon_name(self):
        return "document-new"

class CreateDocumentIn(helplib.reverse_action(CreateNewDocument)):
    rank_adjust = 10
    def __init__(self):
        Action.__init__(self, _("Create Document In..."))

class TemplatesSource (Source, FilesystemWatchMixin):
    def __init__(self):
        Source.__init__(self, _("Document Templates"))

    @classmethod
    def _get_tmpl_dir(self):
        tmpl_dir = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_TEMPLATES)
        if tmpl_dir == os.path.expanduser("~"):
            tmpl_dir = None
        if not tmpl_dir:
            tmpl_dir = os.path.expanduser(DEFAULT_TMPL_DIR)
        pretty.print_debug(__name__, tmpl_dir)
        return tmpl_dir

    def initialize(self):
        self.monitor_token = self.monitor_directories(self._get_tmpl_dir())

    def get_items(self):
        tmpl_dir = self._get_tmpl_dir()
        yield EmptyFile()
        yield NewFolder()
        try:
            for fname in os.listdir(tmpl_dir):
                yield Template(os.path.join(tmpl_dir, fname))
        except EnvironmentError as exc:
            self.output_error(exc)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return None
    def get_icon_name(self):
        return "system-file-manager"

    def provides(self):
        yield Template




__kupfer_name__ = _("Getting Things GNOME")
__kupfer_sources__ = ("TasksSource", )
__kupfer_actions__ = ("CreateNewTask", "CreateNewEmptyTask")
__description__ = _("Browse and create new tasks in GTG")
__version__ = "2017.2"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>, US"

'''
Changes:
    2012-06-21 Karol Będkowski:
        * support new dbus api introduced in GTG 0.2.9
'''

import os

import dbus

from kupfer import plugin_support
from kupfer import pretty
from kupfer import textutils
from kupfer.objects import Leaf, Action, Source
from kupfer.objects import TextLeaf, NotAvailableError, AppLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.weaklib import dbus_signal_connect_weakly

plugin_support.check_dbus_connection()

GTG_ID = "gtg"
_SERVICE_NAME2 = 'org.gnome.GTG'
_OBJECT_NAME2 = '/org/gnome/GTG'
_IFACE_NAME2 = 'org.gnome.GTG'

_STATUS_RANK = {s: i for i, s in enumerate(['Active', 'Postponed', 'Dismiss', 'Done']) }
_STATUS_RANK_DEFAULT = len(_STATUS_RANK)


def _create_dbus_connection_gtg(iface, obj, service, activate=False):
    ''' Create dbus connection to GTG
        @activate: if True, start program if not running
    '''
    interface = None
    sbus = dbus.SessionBus()
    try:
        proxy_obj = sbus.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus')
        dbus_iface = dbus.Interface(proxy_obj, 'org.freedesktop.DBus')
        if activate or dbus_iface.NameHasOwner(iface):
            obj = sbus.get_object(service, obj)
            if obj:
                interface = dbus.Interface(obj, iface)
    except dbus.exceptions.DBusException as err:
        pretty.print_debug(__name__, err)
    return interface


def _create_dbus_connection(activate=False):
    interface = _create_dbus_connection_gtg(_IFACE_NAME2,
            _OBJECT_NAME2, _SERVICE_NAME2, activate)
    if interface is None:
        pretty.print_error(__name__, 'Cannot connect to GTG via DBus')
        if activate:
            raise NotAvailableError(_("Getting Things GNOME"))
    return interface


def _truncate_long_text(text, maxlen=80):
    if len(text) > maxlen:
        return text[:maxlen - 1] + '…'
    return text


def _load_tasks(interface):
    ''' Load task by dbus interface '''
    tasks = interface.GetTasks()
    if not tasks:
        tasks = interface.GetTasksFiltered("")
    for task in tasks:
        title = task['title'].strip()
        if not title:
            title = task['text'].strip()
        title = _truncate_long_text(title)
        otask = Task(task['id'], title, task['status'])
        otask.duedate = task['duedate']
        otask.startdate = task['startdate']
        otask.tags = task['tags']
        yield otask


def _change_task_status(task_id, status):
    interface = _create_dbus_connection(True)
    task = interface.GetTask(task_id)
    task['status'] = status
    interface.ModifyTask(task_id, task)

class Task (Leaf):
    def __init__(self, task_id, title, status):
        Leaf.__init__(self, task_id, title)
        self.status = status
        self.tags = None
        self.duedate = None
        self.startdate = None

    def get_description(self):
        descr = [self.status]
        if self.duedate:
            descr.append(_("due: %s") % self.duedate)
        if self.startdate:
            descr.append(_("start: %s") % self.startdate)
        if self.tags:
            descr.append(_("tags: %s") % " ".join(self.tags))
        return "  ".join(descr)

    def get_icon_name(self):
        return 'gtg'

    def get_actions(self):
        yield OpenEditor()
        yield Delete()
        if self.status != 'Done':
            yield MarkDone()
        if self.status != 'Dismiss':
            yield Dismiss()

    def sort_key(self):
        return _STATUS_RANK.get(self.status, _STATUS_RANK_DEFAULT)

class OpenEditor (Action):
    rank_adjust = 1
    action_accelerator = "o"

    def __init__(self):
        Action.__init__(self, _("Open"))

    def activate(self, leaf):
        interface = _create_dbus_connection(True)
        interface.OpenTaskEditor(leaf.object)

    def get_icon_name(self):
        return 'document-open'

    def get_description(self):
        return _("Open task in Getting Things GNOME!")


class Delete (Action):
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, _("Delete"))

    def activate(self, leaf):
        interface = _create_dbus_connection(True)
        interface.DeleteTask(leaf.object)

    def get_icon_name(self):
        return 'edit-delete'

    def get_description(self):
        return _("Permanently remove this task")


class MarkDone (Action):
    action_accelerator = "d"
    def __init__(self):
        Action.__init__(self, _("Mark Done"))

    def activate(self, leaf):
        _change_task_status(leaf.object, 'Done')

    def get_icon_name(self):
        return 'gtk-yes'

    def get_description(self):
        return _("Mark this task as done")


class Dismiss (Action):
    action_accelerator = "i"
    def __init__(self):
        Action.__init__(self, _("Dismiss"))

    def activate(self, leaf):
        _change_task_status(leaf.object, 'Dismiss')

    def get_icon_name(self):
        return 'gtk-cancel'

    def get_description(self):
        return _("Mark this task as not to be done anymore")


class CreateNewTask (Action):
    def __init__(self):
        Action.__init__(self, _("Create Task"))

    def activate(self, leaf):
        interface = _create_dbus_connection(True)
        title, body = textutils.extract_title_body(leaf.object)
        interface.OpenNewTask(title, body)

    def item_types(self):
        yield TextLeaf

    def get_icon_name(self):
        return 'document-new'

    def get_description(self):
        return _("Create new task in Getting Things GNOME")

class CreateNewEmptyTask (Action):
    def __init__(self):
        Action.__init__(self, _("Create Task"))

    def activate(self, leaf):
        interface = _create_dbus_connection(True)
        interface.OpenNewTask("", "")

    def item_types(self):
        yield AppLeaf

    def valid_for_item(self, leaf):
        return leaf.get_id() == GTG_ID

    def get_icon_name(self):
        return 'document-new'

    def get_description(self):
        return _("Create new task in Getting Things GNOME")


class TasksSource (AppLeafContentMixin, Source):
    appleaf_content_id = GTG_ID

    def __init__(self, name=None):
        Source.__init__(self, name or __kupfer_name__)
        self._tasks = []
        self._version = 3

    def initialize(self):
        bus = dbus.SessionBus()
        dbus_signal_connect_weakly(bus, "TaskAdded", self._on_tasks_updated,
                                   dbus_interface=_IFACE_NAME2)
        dbus_signal_connect_weakly(bus, "TaskModified", self._on_tasks_updated,
                                   dbus_interface=_IFACE_NAME2)
        dbus_signal_connect_weakly(bus, "TaskDeleted", self._on_tasks_updated,
                                   dbus_interface=_IFACE_NAME2)
        dbus_signal_connect_weakly(bus, "NameOwnerChanged", self._name_owner_changed,
                                   dbus_interface="org.freedesktop.DBus",
                                   arg0=_SERVICE_NAME2)

    def _name_owner_changed(self, name, old, new):
        if new and not self._tasks:
            self.mark_for_update()

    def _on_tasks_updated(self, task_id):
        self.mark_for_update()

    def get_items(self):
        interface = _create_dbus_connection()
        if interface is not None:
            self._tasks = list(_load_tasks(interface))
            self._tasks.sort(key=Task.sort_key)
        return self._tasks

    def get_icon_name(self):
        return 'gtg'

    def provides(self):
        yield Task



__kupfer_name__ = "Core"
# The following attributes are updated later
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_contents__ = ()
__kupfer_actions__ = (
    "SearchInside",
    "CopyToClipboard",
    "Rescan",
    )
__description__ = "Core actions and items"
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

from gi.repository import Gtk, Gdk

from kupfer.objects import Leaf, Action
from kupfer.obj.sources import MultiSource
from kupfer import objects
from kupfer.obj.base import InvalidLeafError
from kupfer import interface
from kupfer import pretty
from kupfer import task

def _is_debug():
    # Return True if Kupfer is in debug mode
    return pretty.debug

def register_subplugin(module):
    attrs = (
        "__kupfer_sources__",
        "__kupfer_actions__",
        "__kupfer_text_sources__",
        "__kupfer_contents__"
    )
    for attr in attrs:
        object_names = getattr(module, attr, ())
        globals()[attr] += object_names
        globals().update((sym, getattr(module, sym)) for sym in object_names)

from kupfer.plugin.core import contents, text, internal, commands

register_subplugin(contents)
register_subplugin(text)
register_subplugin(internal)
register_subplugin(commands)

if _is_debug():
    from kupfer.plugin.core import debug
    register_subplugin(debug)

def initialize_plugin(x):
    from kupfer.plugin.core import alternatives
    alternatives.initialize_alternatives(__name__)


class _MultiSource (MultiSource):
    def is_dynamic(self):
        return False

class SearchInside (Action):
    """Return the content source for a Leaf"""
    def __init__(self):
        super(SearchInside, self).__init__(_("Search Contents"))

    def is_factory(self):
        return True
    def activate(self, leaf):
        if not leaf.has_content():
            raise InvalidLeafError("Must have content")
        return leaf.content_source()

    def activate_multiple(self, objects):
        return _MultiSource([L.content_source() for L in objects])

    def item_types(self):
        yield Leaf
    def valid_for_item(self, leaf):
        return leaf.has_content()

    def get_description(self):
        return _("Search inside this catalog")
    def get_icon_name(self):
        return "kupfer-search"

class CopyToClipboard (Action):
    # rank down since it applies everywhere
    rank_adjust = -2
    def __init__(self):
        Action.__init__(self, _("Copy"))
    def wants_context(self):
        return True
    def activate(self, leaf, ctx):
        clip = Gtk.Clipboard.get_for_display(
                ctx.environment.get_screen().get_display(),
                Gdk.SELECTION_CLIPBOARD)
        interface.copy_to_clipboard(leaf, clip)
    def item_types(self):
        yield Leaf
    def valid_for_item(self, leaf):
        try:
            return bool(interface.get_text_representation(leaf))
        except AttributeError:
            pass
    def get_description(self):
        return _("Copy to clipboard")
    def get_icon_name(self):
        return "edit-copy"


class RescanActionTask(task.ThreadTask):
    def __init__(self, source, async_token, retval):
        task.ThreadTask.__init__(self)
        self.source = source
        self.async_token = async_token
        self.retval = retval

    def thread_do(self):
        self.source.get_leaves(force_update=True)

    def thread_finish(self):
        self.async_token.register_late_result(self.retval)

class Rescan (Action):
    """A source action: Rescan a source!  """
    rank_adjust = -5
    def __init__(self):
        Action.__init__(self, _("Rescan"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        if not leaf.has_content():
            raise InvalidLeafError("Must have content")
        source = leaf.content_source()
        return RescanActionTask(source, ctx, leaf)

    def is_async(self):
        return True

    def get_description(self):
        return _("Force reindex of this source")

    def get_icon_name(self):
        return "view-refresh"

    def item_types(self):
        yield objects.AppLeaf
        yield objects.SourceLeaf

    def valid_for_item(self, item):
        if not item.has_content():
            return False
        if item.content_source().is_dynamic():
            return False
        return _is_debug() or item.content_source().source_user_reloadable



from gi.repository import Gtk

from kupfer.objects import Source, RunnableLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer import pretty
from kupfer import kupferui
from kupfer.version import DESKTOP_ID

__kupfer_sources__ = ("KupferSource", )
__kupfer_actions__ = ()
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

__all__ = __kupfer_sources__ + __kupfer_actions__

def _is_debug():
    # Return True if Kupfer is in debug mode
    return pretty.debug

class DebugRestart (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Restart Kupfer"))

    @classmethod
    def _exec_new_kupfer(cls, executable, argv):
        import os
        os.execvp(executable, [executable] + argv)

    def run(self):
        import atexit
        import sys
        Gtk.main_quit()
        atexit.register(self._exec_new_kupfer, sys.executable, sys.argv)

    def get_description(self):
        return str(self)

    def get_icon_name(self):
        return "view-refresh"

class Quit (RunnableLeaf):
    qf_id = "quit"
    def __init__(self, name=None):
        if not name: name = _("Quit")
        super(Quit, self).__init__(name=name)
    def run(self):
        Gtk.main_quit()
    def get_description(self):
        return _("Quit Kupfer")
    def get_icon_name(self):
        return "application-exit"

class About (RunnableLeaf):
    def __init__(self, name=None):
        if not name: name = _("About Kupfer")
        super(About, self).__init__(name=name)
    def wants_context(self):
        return True
    def run(self, ctx):
        kupferui.show_about_dialog(ctx.environment)
    def get_description(self):
        return _("Show information about Kupfer authors and license")
    def get_icon_name(self):
        return "help-about"

class Help (RunnableLeaf):
    def __init__(self, name=None):
        if not name: name = _("Kupfer Help")
        super(Help, self).__init__(name=name)
    def wants_context(self):
        return True
    def run(self, ctx):
        kupferui.show_help(ctx.environment)
    def get_description(self):
        return _("Get help with Kupfer")
    def get_icon_name(self):
        return "help-contents"

class Preferences (RunnableLeaf):
    def __init__(self, name=None):
        if not name: name = _("Kupfer Preferences")
        super(Preferences, self).__init__(name=name)
    def wants_context(self):
        return True
    def run(self, ctx):
        kupferui.show_preferences(ctx.environment)
    def get_description(self):
        return _("Show preferences window for Kupfer")
    def get_icon_name(self):
        return "preferences-desktop"

class KupferSource (AppLeafContentMixin, Source):
    appleaf_content_id = DESKTOP_ID
    def __init__(self, name=_("Kupfer")):
        Source.__init__(self, name)
    def is_dynamic(self):
        return True
    def get_items(self):
        yield Preferences()
        yield Help()
        yield About()
        yield Quit()
        yield DebugRestart()

    def get_description(self):
        return None
    def get_icon_name(self):
        return "search"
    def provides(self):
        yield RunnableLeaf




__kupfer_name__ = _("Send Keys")
__kupfer_actions__ = (
    "CopyAndPaste",
    "SendKeys",
    "TypeText",
    )
__description__ = _("Send synthetic keyboard events using "
                    "xautomation")
__version__ = ""
__author__ = ""

from gi.repository import Gtk, Gdk

from kupfer.objects import Leaf, Action, TextLeaf
from kupfer.objects import OperationError
from kupfer import utils
from kupfer import interface

# delay for first keypress and all following
INIT_DELAY = 'usleep 300000'
INTER_DELAY = 'usleep 50000'

class CopyAndPaste (Action):
    # rank down since it applies everywhere
    rank_adjust = -2
    def __init__(self):
        Action.__init__(self, _("Paste to Foreground Window"))
    def activate(self, leaf):
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        interface.copy_to_clipboard(leaf, clip)
        xte_paste_argv = ['xte', INIT_DELAY, 'keydown Control_L',
                          'key v', 'keyup Control_L']
        try:
            utils.spawn_async_raise(xte_paste_argv)
        except utils.SpawnError as exc:
            raise OperationError(exc)
    def item_types(self):
        yield Leaf
    def valid_for_item(self, leaf):
        try:
            return bool(interface.get_text_representation(leaf))
        except AttributeError:
            pass
    def get_description(self):
        return _("Copy to clipboard and send Ctrl+V to foreground window")
    def get_icon_name(self):
        return "edit-paste"

class SendKeys (Action):
    def __init__(self):
        Action.__init__(self, _("Send Keys"))

    def activate(self, leaf):
        return self.activate_multiple((leaf, ))

    def activate_multiple(self, objects):
        xte_sendkey_argv = ['xte', INIT_DELAY]
        iterobjects = iter(objects)
        for obj in iterobjects:
            xte_sendkey_argv.extend(self.make_keystr_arguments(obj.object))
            break
        for obj in iterobjects:
            xte_sendkey_argv.append(INTER_DELAY)
            xte_sendkey_argv.extend(self.make_keystr_arguments(obj.object))

        try:
            utils.spawn_async_raise(xte_sendkey_argv)
        except utils.SpawnError as exc:
            raise OperationError(exc)

    def make_keystr_arguments(self, keystr):
        keys, orig_mods = Gtk.accelerator_parse(keystr)
        m = {
            Gdk.ModifierType.SHIFT_MASK: "Shift_L",
            Gdk.ModifierType.CONTROL_MASK: "Control_L",
            Gdk.ModifierType.SUPER_MASK: "Super_L",
            Gdk.ModifierType.MOD1_MASK: "Alt_L",
        }
        mod_names = []
        mods = orig_mods
        for mod in m:
            if mod & mods:
                mod_names.append(m[mod])
                mods &= ~mod
        if mods != 0:
            raise OperationError("Keys not yet implemented: %s" %
                    Gtk.accelerator_get_label(keys, orig_mods))
        key_arg = 'key %s' % (Gdk.keyval_name(keys), )
        mods_down = ['keydown ' + n for n in mod_names]
        mods_up = ['keyup ' + n for n in reversed(mod_names)]
        return mods_down + [key_arg] + mods_up

    def item_types(self):
        yield TextLeaf
    def valid_for_item(self, leaf):
        text = leaf.object
        keys, mods = Gtk.accelerator_parse(text)
        return keys
    def get_description(self):
        return _("Send keys to foreground window")

class TypeText (Action):
    rank_adjust = -2 
    def __init__(self):
        Action.__init__(self, _("Type Text"))
    def activate(self, leaf):
        text = interface.get_text_representation(leaf)
        xte_paste_argv = ['xte', 'usleep 300000']
        # replace all newlines with 'key Return'
        for line in text.splitlines(True):
            xte_paste_argv.append("str " + line.rstrip("\r\n"))
            if line.endswith("\n"):
                xte_paste_argv.append("key Return")
        try:
            utils.spawn_async_raise(xte_paste_argv)
        except utils.SpawnError as exc:
            raise OperationError(exc)
    def item_types(self):
        yield Leaf
    def valid_for_item(self, leaf):
        try:
            return bool(interface.get_text_representation(leaf))
        except AttributeError:
            pass
    def get_description(self):
        return _("Type the text to foreground window")





ACCELERATOR_NAMES = {
    # TRANS: Names of accelerators in the interface
    'activate': _('Alternate Activate'),
    # TRANS: The "Comma Trick"/"Put Selection on Stack" allows the
    # TRANS: user to select many objects to be used for one action
    'comma_trick': _('Comma Trick'),
    # TRANS: "Compose Command" makes one object out of the selected
    # TRANS: object + action (+iobject)
    'compose_action': _('Compose Command'),
    'mark_as_default': _('Mark Default Action'),
    'erase_affinity_for_first_pane': _('Forget Object'),
    'reset_all': _('Reset All'),
    'select_quit': _('Select Quit'),
    'select_selected_file': _('Select Selected File'),
    'select_selected_text': _('Select Selected Text'),
    'select_clipboard_file': _('Select Clipboard File'),
    'select_clipboard_text': _('Select Clipboard Text'),
    'show_help': _('Show Help'),
    'show_preferences': _('Show Preferences'),
    'switch_to_source': _('Switch to 1st Pane'),
    'switch_to_2': _('Switch to 2nd Pane'),
    'switch_to_3': _('Switch to 3rd Pane'),
    "toggle_text_mode_quick": _('Toggle Text Mode'),
}



__kupfer_name__ = _("Prefer Dark Theme")
__description__ = ""
__version__ = "2017.1"
__author__ = ""

from gi.repository import Gtk

from kupfer import plugin_support, pretty

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "prefer_dark",
        "label": _("Prefer Dark Theme"),
        "type": bool,
        "value": True,
    },
)

def initialize_plugin(name):
    use_theme(__kupfer_settings__['prefer_dark'])
    __kupfer_settings__.connect_settings_changed_cb(on_change_theme)

def finalize_plugin(name):
    use_theme(None)

def on_change_theme(sender, key, value):
    use_theme(value)

PREFER_DARK = "gtk-application-prefer-dark-theme"

def use_theme(enabled):
    pretty.print_debug(__name__, "updating setting to", enabled)
    s = Gtk.Settings.get_default()
    if enabled is None:
        s.reset_property(PREFER_DARK)
    else:
        s.set_property(PREFER_DARK, enabled)



import contextlib

from kupfer import pretty

from kupfer.core import plugins
from kupfer.core.plugins import (load_plugin_sources, sources_attribute,
        action_decorators_attribute, text_sources_attribute,
        content_decorators_attribute, action_generators_attribute,
        initialize_plugin)

class PluginDescription (object):
    text_sources = ()
    action_decorators = ()
    content_decorators = ()
    action_generators = ()
    sources = ()

def load_plugin(plugin_id):
    """
    @S_sources are to be included directly in the catalog,
    @s_souces as just as subitems
    """
    sources = []
    text_sources = []
    action_decorators = []
    content_decorators = []
    action_generators = []

    item = plugin_id

    initialize_plugin(item)
    if not plugins.is_plugin_loaded(item):
        return PluginDescription()
    text_sources.extend(load_plugin_sources(item, text_sources_attribute))
    action_decorators.extend(load_plugin_sources(item,
        action_decorators_attribute))
    action_generators.extend(load_plugin_sources(item,
        action_generators_attribute))

    # Register all Sources as (potential) content decorators
    content_decorators.extend(load_plugin_sources(item,
        sources_attribute, instantiate=False))
    content_decorators.extend(load_plugin_sources(item,
        content_decorators_attribute, instantiate=False))
    sources.extend(load_plugin_sources(item))

    desc = PluginDescription()

    desc.text_sources = text_sources
    desc.action_decorators = action_decorators
    desc.content_decorators = content_decorators
    desc.sources = sources
    desc.action_generators = action_generators
    return desc

@contextlib.contextmanager
def exception_guard(name, callback=None, *args, **kwargs):
    "Guard for exceptions, print traceback and call @callback if any is raised"
    try:
        yield
    except Exception:
        import traceback
        pretty.print_error(__name__, "Loading %s raised an exception:" % name)
        traceback.print_exc()
        pretty.print_error(__name__, "This error is probably a bug in", name)
        pretty.print_error(__name__, "Please file a bug report")
        if callback is not None:
            callback(*args, **kwargs)

def remove_plugin(plugin_id):
    plugins.unimport_plugin(plugin_id)



# -*- coding: UTF-8 -*-
'''
virtualbox_vboxapi_support.py

Control VirtualBox via Python interface (vboxapi).
Only (?) Sun VirtualBox (no OSE).
'''
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = "2018-10-21"

import vboxapi

from kupfer import pretty


# check api
try:
    vboxapi.VirtualBoxReflectionInfo(None).SessionState_Locked
except AttributeError:
    raise ImportError()

from kupfer.plugin.virtualbox import constants as vbox_const

MONITORED_DIRS = None
IS_DYNAMIC = False
ICON = "VBox"
APP_ID = "virtualbox"


_ACTIONS = {
    vbox_const.VM_POWEROFF: lambda c: c.powerDown(),
    vbox_const.VM_ACPI_POWEROFF: lambda c: c.powerButton(),
    vbox_const.VM_PAUSE: lambda c: c.pause(),
    vbox_const.VM_REBOOT: lambda c: c.reset(),
    vbox_const.VM_RESUME: lambda c: c.resume(),
    vbox_const.VM_SAVE: lambda c: c.saveState(),
}


_VBOX = vboxapi.VirtualBoxManager(None, None)


def get_machine_by_id(mid):
    try:
        mach = _VBOX.getVirtualBox().getMachine(mid)
    except AttributeError:
        mach = _VBOX.getVirtualBox().findMachine(mid)
    return mach


def get_machine_state(machine_id):
    ''' check vms state (on/off/paused) '''
    if _VBOX is None:
        return vbox_const.VM_STATE_POWEROFF
    state = vbox_const.VM_STATE_POWERON
    machine_state = None
    try:
        machine = get_machine_by_id(machine_id)
        pretty.print_debug(__name__, "machine ", repr(machine))
        machine_state = machine.state
        if machine_state == _VBOX.constants.MachineState_Paused:
            state = vbox_const.VM_STATE_PAUSED
        elif machine_state in (_VBOX.constants.MachineState_PoweredOff,
                               _VBOX.constants.MachineState_Aborted,
                               _VBOX.constants.MachineState_Starting):
            state = vbox_const.VM_STATE_POWEROFF
        elif machine_state == _VBOX.constants.MachineState_Saved:
            state = vbox_const.VM_STATE_SAVED
    except Exception as err:  # exception == machine is off (xpcom.Exception)
        pretty.print_debug(__name__, 'get_machine_state', machine_state, err)
        # silently set state to off
        state = vbox_const.VM_STATE_POWEROFF
    return state


def _machine_start(vm_uuid, mode):
    ''' Start virtual machine
        @param vm_uuid - uuid of virtual machine
        @param mode - mode: gui, headless
    '''
    try:
        session = _VBOX.getSessionObject()
        mach = get_machine_by_id(vm_uuid)
        remote_sess = mach.launchVMProcess(session, mode, '')
        remote_sess.waitForCompletion(-1)
        session.unlockMachine()
    except Exception as err:
        pretty.print_error(__name__, "StartVM:", vm_uuid, "Mode ", mode,
                           "error", err)


def _execute_machine_action(vm_uuid, action):
    ''' Start virtual machine
        @param vm_uuid - uuid of virtual machine
        @param action - function called on _VBOX session
    '''
    try:
        session = _VBOX.getSessionObject()
        mach = get_machine_by_id(vm_uuid)
        mach.lockMachine(session, _VBOX.constants.LockType_Shared)
        action(session.console)
        session.unlockMachine()
    except Exception as err:
        pretty.print_error(__name__, "_execute_machine_action:", repr(action),
                           " vm:", vm_uuid, "error", err)


def vm_action(action, vm_uuid):
    ''' change state of the virtual machine
        @param action - one of the const VM_*
        @param vm_uuid - virtual machine uuid
    '''
    if _VBOX is None:
        return
    if action == vbox_const.VM_START_NORMAL:
        _machine_start(vm_uuid, 'gui')
    elif action == vbox_const.VM_START_HEADLESS:
        _machine_start(vm_uuid, 'headless')
    else:
        command = _ACTIONS[action]
        _execute_machine_action(vm_uuid, command)


def get_machines():
    ''' Get generator of items:
        (machine uuid, machine name, machine description)
    '''
    if _VBOX is None:
        return

    machines = _VBOX.getArray(_VBOX.getVirtualBox(), 'machines')
    for machine in machines:
        if not machine.accessible:
            continue
        description = machine.description or machine.OSTypeId
        yield (machine.id, machine.name, description)


def unload():
    pass



__kupfer_name__ = _("URL Actions")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
        "DownloadAndOpen",
        "DownloadTo",
    )
__description__ = _("URL Actions")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import shutil
import urllib.request, urllib.parse, urllib.error

from kupfer.objects import Action, UrlLeaf, FileLeaf
from kupfer import utils, task

class DownloadTask (task.ThreadTask):
    def __init__(self, uri, destdir=None, tempfile=False, finish_callback=None):
        super(DownloadTask, self).__init__()
        self.uri = uri
        self.download_finish_callback = finish_callback
        self.destdir = destdir
        self.use_tempfile = tempfile

    def thread_do(self):
        self.response = urllib.request.urlopen(self.uri)

        def url_name(url):
            return os.path.basename(url.rstrip("/"))
        def header_name(headers):
            content_disp = headers.get("Content-Disposition", "")
            for part in content_disp.split(";"):
                if part.strip().lower().startswith("filename="):
                    return part.split("=", 1)[-1]
            return content_disp

        destname = (header_name(self.response.headers) or
                    url_name(self.response.url))

        if self.use_tempfile:
            (self.destfile, self.destpath) = utils.get_safe_tempfile()
        else:
            (self.destfile, self.destpath) = \
                utils.get_destfile_in_directory(self.destdir, destname)
        try:
            if not self.destfile:
                raise IOError("Could not write output file")

            shutil.copyfileobj(self.response, self.destfile)
        finally:
            self.destfile.close()
            self.response.close()

    def thread_finish(self):
        if self.download_finish_callback:
            self.download_finish_callback(self.destpath)

class DownloadAndOpen (Action):
    """Asynchronous action to download file and open it"""
    def __init__(self):
        Action.__init__(self, _("Download and Open"))

    def is_async(self):
        return True
    def wants_context(self):
        return True
    def activate(self, leaf, ctx):
        uri = leaf.object
        def finish_action(filename):
            utils.show_path(filename)
            ctx.register_late_result(FileLeaf(filename), show=False)
        return DownloadTask(uri, None, True, finish_action)

    def item_types(self):
        yield UrlLeaf
    def get_description(self):
        return None

class DownloadTo (Action):
    def __init__(self):
        Action.__init__(self, _("Download To..."))

    def is_async(self):
        return True
    def wants_context(self):
        return True
    def activate(self, leaf, obj, ctx):
        uri = leaf.object
        def finish_action(filename):
            ctx.register_late_result(FileLeaf(filename))
        return DownloadTask(uri, obj.object, False, finish_action)

    def item_types(self):
        yield UrlLeaf
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item=None):
        return utils.is_directory_writable(obj.object)
    def get_description(self):
        return _("Download URL to a chosen location")




# -*- coding: UTF-8 -*-

__kupfer_name__ = _("Deep Directories")
__kupfer_sources__ = ("DeepDirSource",)
__description__ = _("Recursive index directories")
__version__ = "2018-09-04"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

'''
Changes:
    2018-09-04 * fix: Python 3 compatibility
	2012-10-08 * fix: errors when no one configured directories exists
	2012-06-09 + max depth; fix source name
	2012-06-08 - init
'''

import os

from kupfer import plugin_support
from kupfer.obj import sources

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        'key': 'dirs',
        'label': _("Directories (;-separated):"),
        'type': str,
        'value': "~/Documents/",
    },
    {
        'key': 'depth',
        'label': _("Depth (max 10):"),
        'type': int,
        'value': 2,
    },
)

MAX_DEPTH = 10


class DeepDirSource(sources.FileSource):
    def __init__(self, name=_("Deep Directories")):
        sources.FileSource.__init__(self, self._get_dirs() or [''],
                                    min(__kupfer_settings__['depth'], MAX_DEPTH))
        self.name = name

    def initialized(self):
        __kupfer_settings__.connect("plugin-setting-changed",
                                    self._setting_changed)

    def get_items(self):
        self.dirlist = self._get_dirs()
        if not self.dirlist:
            return []
        self.depth = min(__kupfer_settings__['depth'], MAX_DEPTH)
        return sources.FileSource.get_items(self)

    @staticmethod
    def _get_dirs():
        if not __kupfer_settings__['dirs']:
            return []
        return list(filter(os.path.isdir,
                           (os.path.expanduser(path)
                            for path
                            in __kupfer_settings__['dirs'].split(';'))))

    def _setting_changed(self, settings, key, value):
        if key in ('dirs', 'depth'):
            self.mark_for_update()



__kupfer_name__ = _("Custom Terminal")
__description__ = _("""Configure a custom terminal emulator.

The plugin adds another terminal alternative (select it on the main page).""")
__version__ = ""
__author__ = "Ulrik Sverdrup"

from kupfer import plugin_support
from kupfer import utils

__kupfer_settings__ = plugin_support.PluginSettings(
        {
            "key": "command",
            "label": _("Command"),
            "type": str,
            "value": "",
        },
        {
            "key": "exearg",
            "label": _("Execute flag"),
            "type": str,
            "value": "-e",
            "tooltip": ("The flag which makes the terminal execute"
                        " everything following it in the argument string. ")
        },
    )



def initialize_plugin(name):
    __kupfer_settings__.connect_settings_changed_cb(_update_alternative)
    _update_alternative()

def _update_alternative(*args):
    command = __kupfer_settings__["command"]
    exearg = __kupfer_settings__["exearg"]
    argv = utils.argv_for_commandline(command)
    if not argv or not utils.lookup_exec_path(argv[0]):
        return
    plugin_support.register_alternative(__name__, 'terminal', 'custom1',
            name=_("Custom Terminal"),
            argv=argv,
            exearg=exearg,
            desktopid="",
            startup_notify=True)





from kupfer.objects import Source, Leaf
from kupfer.objects import RunnableLeaf
from kupfer.core import commandexec

__kupfer_sources__ = ("KupferInterals", "CommandResults", )
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

class LastCommand (RunnableLeaf):
    "Represented object is the command tuple to run"
    qf_id = "lastcommand"
    def __init__(self, obj):
        RunnableLeaf.__init__(self, obj, _("Last Command"))

    def wants_context(self):
        return True

    def run(self, ctx):
        obj, action, iobj = self.object
        return ctx.delegated_run(obj, action, iobj)

class KupferInterals (Source):
    def __init__(self):
        Source.__init__(self, _("Internal Kupfer Objects"))

    def is_dynamic(self):
        return True

    def get_items(self):
        ctx = commandexec.DefaultActionExecutionContext()
        if ctx.last_command is None:
            return
        yield LastCommand(ctx.last_command)

    def provides(self):
        yield LastCommand

class LastResultObject (Leaf):
    "dummy superclass"

def _make_first_result_object(leaf):
    global LastResultObject
    class LastResultObject (LastResultObject):
        qf_id = "lastresult"
        def __init__(self, leaf):
            Leaf.__init__(self, leaf.object, _("Last Result"))
            vars(self).update(vars(leaf))
            self.name = _("Last Result")
            self.__orignal_leaf = leaf
            self.__class__.__bases__ = (leaf.__class__, Leaf)

        def get_gicon(self):
            return None
        def get_icon_name(self):
            return Leaf.get_icon_name(self)
        def get_thumbnail(self, w, h):
            return None
        def get_description(self):
            return str(self.__orignal_leaf)

    return LastResultObject(leaf)


class CommandResults (Source):
    def __init__(self):
        Source.__init__(self, _("Command Results"))

    def is_dynamic(self):
        return True

    def get_items(self):
        ctx = commandexec.DefaultActionExecutionContext()
        for x in reversed(ctx.last_results):
            yield x
        try:
            leaf = ctx.last_results[-1]
        except IndexError:
            return
        yield _make_first_result_object(leaf)

    def provides(self):
        yield Leaf
        yield LastResultObject



"""
A test project to see if we can make a plugin that allows us to
drill down into compressed archives.

So far we only support .zip and .tar, .tar.gz, .tar.bz2, using Python's
standard library.
"""
__kupfer_name__ = _("Deep Archives")
__kupfer_contents__ = ("ArchiveContent", )
__description__ = _("Allow browsing inside compressed archive files")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import hashlib
import os
import shutil
import tarfile
import zipfile

from kupfer.objects import Source, FileLeaf
from kupfer.obj.sources import DirectorySource
from kupfer import pretty
from kupfer import scheduler

# Limit this to archives of a couple of megabytes
MAX_ARCHIVE_BYTE_SIZE = 15 * 1024**2

# Wait a year, or until program shutdown for cleaning up
# archive files
VERY_LONG_TIME_S = 3600*24*365


class UnsafeArchiveError (Exception):
    def __init__(self, path):
        Exception.__init__(self, "Refusing to extract unsafe path: %s" % path)

def is_safe_to_unarchive(path):
    "return whether @path is likely a safe path to unarchive"
    npth = os.path.normpath(path)
    return not os.path.isabs(npth) and not npth.startswith(os.path.pardir)


class ArchiveContent (Source):
    extractors = []
    unarchived_files = []
    end_timer = scheduler.Timer(True)

    def __init__(self, fileleaf, unarchive_func):
        Source.__init__(self, _("Content of %s") % fileleaf)
        self.path = fileleaf.object
        self.unarchiver = unarchive_func

    def repr_key(self):
        return self.path

    def get_items(self):
        # always use the same destination for the same file and mtime
        basename = os.path.basename(os.path.normpath(self.path))
        root, ext = os.path.splitext(basename)
        mtime = os.stat(self.path).st_mtime
        fileid = hashlib.sha1(("%s%s" % (self.path, mtime)).encode("utf-8")).hexdigest()
        pth = os.path.join("/tmp", "kupfer-%s-%s" % (root, fileid, ))
        if not os.path.exists(pth):
            self.output_debug("Extracting with %s" % (self.unarchiver, ))
            self.unarchiver(self.path, pth)
            self.unarchived_files.append(pth)
            self.end_timer.set(VERY_LONG_TIME_S, self.clean_up_unarchived_files)
        files = list(DirectorySource(pth, show_hidden=True).get_leaves())
        if len(files) == 1 and files[0].has_content():
            return files[0].content_source().get_leaves()
        return files

    def get_description(self):
        return None

    @classmethod
    def decorates_type(cls):
        return FileLeaf

    @classmethod
    def decorate_item(cls, leaf):
        basename = os.path.basename(leaf.object).lower()
        for extractor in cls.extractors:
            if any(basename.endswith(ext) for ext in extractor.extensions):
                if os.path.isfile(leaf.object) and extractor.predicate(leaf.object):
                    return cls._source_for_path(leaf, extractor)


    @classmethod
    def _source_for_path(cls, leaf, extractor):
        byte_size = os.stat(leaf.object).st_size
        if byte_size < MAX_ARCHIVE_BYTE_SIZE:
            return cls(leaf, extractor)
        return None

    @classmethod
    def clean_up_unarchived_files(cls):
        if not cls.unarchived_files:
            return
        pretty.print_info(__name__, "Removing extracted archives..")
        for filetree in set(cls.unarchived_files):
            pretty.print_debug(__name__, "Removing", os.path.basename(filetree))
            shutil.rmtree(filetree, onerror=cls._clean_up_error_handler)
        cls.unarchived_files = []


    @classmethod
    def _clean_up_error_handler(cls, func, path, exc_info):
        pretty.print_error(__name__, "Error in %s deleting %s:" % (func, path))
        pretty.print_error(__name__, exc_info)

    @classmethod
    def extractor(cls, extensions, predicate):
        def decorator(func):
            func.extensions = extensions
            func.predicate = predicate
            cls.extractors.append(func)
            return func
        return decorator


@ArchiveContent.extractor((".tar", ".tar.gz", ".tgz", ".tar.bz2"),
        tarfile.is_tarfile)
def extract_tarfile(filepath, destpath):
    zf = tarfile.TarFile.open(filepath, 'r')
    try:
        for member in zf.getnames():
            if not is_safe_to_unarchive(member):
                raise UnsafeArchiveError(member)
        zf.extractall(path=destpath)
    finally:
        zf.close()


# ZipFile only supports extractall since Python 2.6
@ArchiveContent.extractor((".zip", ), zipfile.is_zipfile)
def extract_zipfile(filepath, destpath):
    zf = zipfile.ZipFile(filepath, 'r')
    try:
        for member in zf.namelist():
            if not is_safe_to_unarchive(member):
                raise UnsafeArchiveError(member)
        zf.extractall(path=destpath)
    finally:
        zf.close()



__kupfer_name__ = _("Audacious")
__kupfer_sources__ = ("AudaciousSource", )
__description__ = _("Control Audacious playback and playlist")
__version__ = "2017.2"
__author__ = "Horia V. Corcalciuc <h.v.corcalciuc@gmail.com>, US"

import subprocess

from kupfer.objects import Leaf, Source, Action
from kupfer.objects import RunnableLeaf, SourceLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer import icons, utils, uiutils
from kupfer import plugin_support
from kupfer import kupferstring
from kupfer.weaklib import dbus_signal_connect_weakly

plugin_support.check_dbus_connection()

import dbus

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "playlist_toplevel",
        "label": _("Include songs in top level"),
        "type": bool,
        "value": True,
    },
)

AUDTOOL = "audtool"
AUDACIOUS = "audacious"
_BUS_NAME = "org.atheme.audacious"

def enqueue_song(info):
    utils.spawn_async((AUDTOOL, "playqueue-add", "%d" % info))

def dequeue_song(info):
    utils.spawn_async((AUDTOOL, "playqueue-remove", "%d" % info))

def play_song(info):
    utils.spawn_async((AUDTOOL, "playlist-jump", "%d" % info))
    utils.spawn_async((AUDTOOL, "playback-play"))

def get_playlist_songs():
    """Yield tuples of (position, name) for playlist songs"""
    toolProc = subprocess.Popen([AUDTOOL, "playlist-display"],
            stdout=subprocess.PIPE)
    stdout, stderr = toolProc.communicate()
    for line in stdout.splitlines():
        if not line.count(b'|') >= 2:
            continue
        position, rest = line.split(b'|', 1)
        songname, rest = rest.rsplit(b'|', 1)
        pos = int(position.strip())
        nam = kupferstring.fromlocale(songname.strip())
        yield (pos, nam)

def get_current_song():
    toolProc = subprocess.Popen([AUDTOOL, "current-song"],
            stdout=subprocess.PIPE)
    stdout, stderr = toolProc.communicate()
    for line in stdout.splitlines():
        return kupferstring.fromlocale(line)
    return None

def clear_queue():
    utils.spawn_async((AUDTOOL, "playqueue-clear"))

class Enqueue (Action):
    action_accelerator = "e"
    def __init__(self):
        Action.__init__(self, _("Enqueue"))
    def activate(self, leaf):
        enqueue_song(leaf.object)
    def get_description(self):
        return _("Add track to the Audacious play queue")
    def get_gicon(self):
        return icons.ComposedIcon("gtk-execute", "media-playback-start")
    def get_icon_name(self):
        return "media-playback-start"

class Dequeue (Action):
    def __init__(self):
        Action.__init__(self, _("Dequeue"))
    def activate(self, leaf):
        dequeue_song(leaf.object)
    def get_description(self):
        return _("Remove track from the Audacious play queue")
    def get_gicon(self):
        return icons.ComposedIcon("gtk-execute", "media-playback-stop")
    def get_icon_name(self):
        return "media-playback-stop"

class JumpToSong(Action):
    action_accelerator = "o"
    def __init__(self):
        Action.__init__(self, _("Play"))
    def activate(self, leaf):
        play_song(leaf.object)
    def get_description(self):
        return _("Jump to track in Audacious")
    def get_icon_name(self):
        return "media-playback-start"

class Play (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Play"))
    def run(self):
        utils.spawn_async((AUDTOOL, "playback-play"))
    def get_description(self):
        return _("Resume playback in Audacious")
    def get_icon_name(self):
        return "media-playback-start"

class Pause (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Pause"))
    def run(self):
        utils.spawn_async((AUDTOOL, "playback-pause"))
    def get_description(self):
        return _("Pause playback in Audacious")
    def get_icon_name(self):
        return "media-playback-pause"

class Next (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Next"))
    def run(self):
        utils.spawn_async((AUDTOOL, "playlist-advance"))
    def get_description(self):
        return _("Jump to next track in Audacious")
    def get_icon_name(self):
        return "media-skip-forward"

class Previous (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Previous"))
    def run(self):
        utils.spawn_async((AUDTOOL, "playlist-reverse"))
    def get_description(self):
        return _("Jump to previous track in Audacious")
    def get_icon_name(self):
        return "media-skip-backward"

class ClearQueue (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Clear Queue"))
    def run(self):
        clear_queue()
    def get_description(self):
        return _("Clear the Audacious play queue")
    def get_icon_name(self):
        return "edit-clear"
        
class Shuffle (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Shuffle"))
    def run(self):
        utils.spawn_async((AUDTOOL, "playlist-shuffle-toggle"))
    def get_description(self):
        return _("Toggle shuffle in Audacious")
    def get_icon_name(self):
        return "media-playlist-shuffle"

class Repeat (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Repeat"))
    def run(self):
        utils.spawn_async((AUDTOOL, "playlist-repeat-toggle"))
    def get_description(self):
        return _("Toggle repeat in Audacious")
    def get_icon_name(self):
        return "media-playlist-repeat"

class ShowPlaying (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Show Playing"))
    def run(self):
        song = get_current_song()
        if song is not None:
            uiutils.show_notification(song, icon_name="audio-x-generic")
    def get_description(self):
        return _("Tell which song is currently playing")
    def get_gicon(self):
        return icons.ComposedIcon("dialog-information", "audio-x-generic")
    def get_icon_name(self):
        return "dialog-information"

class SongLeaf (Leaf):
    """The SongLeaf's represented object is the Playlist index"""
    def get_actions(self):
        yield JumpToSong()
        yield Enqueue()
        yield Dequeue()
    def get_icon_name(self):
        return "audio-x-generic"

class AudaciousSongsSource (Source):
    def __init__(self, library):
        Source.__init__(self, _("Playlist"))
        self.library = library
    def get_items(self):
        for song in self.library:
            yield SongLeaf(*song)
    def get_gicon(self):
        return icons.ComposedIcon(AUDACIOUS, "audio-x-generic",
            emblem_is_fallback=True)
    def provides(self):
        yield SongLeaf

class AudaciousSource (AppLeafContentMixin, Source):
    appleaf_content_id = AUDACIOUS
    source_user_reloadable = True

    def __init__(self):
        Source.__init__(self, _("Audacious"))

    def initialize(self):
        bus = dbus.SessionBus()
        dbus_signal_connect_weakly(bus, "NameOwnerChanged", self._name_owner_changed,
                                   dbus_interface="org.freedesktop.DBus",
                                   arg0=_BUS_NAME)

    def _name_owner_changed(self, name, old, new):
        if new:
            self.mark_for_update()

    def get_items(self):
        yield Play()
        yield Pause()
        yield Next()
        yield Previous() 
        yield ClearQueue()
        yield ShowPlaying()
        # Commented as these seem to have no effect
        #yield Shuffle()
        #yield Repeat()
        songs = list(get_playlist_songs())
        songs_source = AudaciousSongsSource(songs)
        yield SourceLeaf(songs_source)
        if __kupfer_settings__["playlist_toplevel"]:
            for leaf in songs_source.get_leaves():
                yield leaf
    def get_description(self):
        return __description__
    def get_icon_name(self):
        return AUDACIOUS
    def provides(self):
        yield RunnableLeaf



import gi
from gi.repository import GObject

from kupfer import pretty
from kupfer import environment

KEYBINDING_DEFAULT = 1
KEYBINDING_MAGIC = 2

KEYRANGE_RESERVED = (3, 0x1000)
KEYRANGE_TRIGGERS = (0x1000, 0x2000)

Keybinder = None
if environment.allows_keybinder():
    try:
        gi.require_version("Keybinder", "3.0")
    except ValueError:
        pretty.print_debug(__name__, "Keybinder 3.0 not available in gi")
    else:
        from gi.repository import Keybinder
        Keybinder.init()
else:
    pretty.print_debug(__name__, "Keybinder disabled")

_keybound_object = None
def GetKeyboundObject():
    """Get the shared instance"""
    global _keybound_object
    if not _keybound_object:
        _keybound_object = KeyboundObject()
    return _keybound_object

class KeyboundObject (GObject.GObject):
    """Keybinder object

    signals:
        keybinding (target, event_time)
        keybinding signal is triggered when the key bound
        for @target is triggered.
    """
    __gtype_name__ = "KeyboundObject"
    def __init__(self):
        super(KeyboundObject, self).__init__()
    def _keybinding(self, target):
        time = Keybinder.get_current_event_time()
        self.emit("keybinding", target, "", time)
    def emit_bound_key_changed(self, keystring, is_bound):
        self.emit("bound-key-changed", keystring, is_bound)
    def relayed_keys(self, sender, keystring, display, timestamp):
        for target, key in _currently_bound.items():
            if keystring == key:
                self.emit("keybinding", target, display, timestamp)

# Arguments: Target, Display, Timestamp
GObject.signal_new("keybinding", KeyboundObject, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_INT, GObject.TYPE_STRING, GObject.TYPE_UINT))
# Arguments: Keystring, Boolean
GObject.signal_new("bound-key-changed", KeyboundObject, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_BOOLEAN,))

_currently_bound = {}

def is_available():
    """
    Return True if keybindings are available.
    """
    if Keybinder is None:
        return False
    else:
        try:
            return Keybinder.supported()
        except AttributeError:
            return True


def get_all_bound_keys():
    return list(filter(bool, list(_currently_bound.values())))

def get_current_event_time():
    "Return current event time as given by keybinder"
    if Keybinder is None:
        return 0
    return Keybinder.get_current_event_time()

def _register_bound_key(keystr, target):
    _currently_bound[target] = keystr

def get_currently_bound_key(target=KEYBINDING_DEFAULT):
    return _currently_bound.get(target)

def bind_key(keystr, keybinding_target=KEYBINDING_DEFAULT):
    """
    Bind @keystr, unbinding any previous key for @keybinding_target.
    If @keystr is a false value, any previous key will be unbound.
    """
    keybinding_target = int(keybinding_target)

    if Keybinder is None:
        return False

    def callback(keystr):
        return GetKeyboundObject()._keybinding(keybinding_target)

    if not _is_sane_keybinding(keystr):
        pretty.print_error(__name__, "Refusing to bind key", repr(keystr))
        return False

    succ = True
    if keystr:
        try:
            succ = Keybinder.bind(keystr, callback)
            pretty.print_debug(__name__, "binding", repr(keystr))
            GetKeyboundObject().emit_bound_key_changed(keystr, True)
        except KeyError as exc:
            pretty.print_error(__name__, exc)
            succ = False
    if succ:
        old_keystr = get_currently_bound_key(keybinding_target)
        if old_keystr and old_keystr != keystr:
            Keybinder.unbind(old_keystr)
            pretty.print_debug(__name__, "unbinding", repr(old_keystr))
            GetKeyboundObject().emit_bound_key_changed(old_keystr, False)
        _register_bound_key(keystr, keybinding_target)
    return succ


def _is_sane_keybinding(keystr):
    "Refuse keys that we absolutely do not want to bind"
    if keystr is None:
        return True
    if len(keystr) == 1 and keystr.isalnum():
        return False
    if keystr in set(["Return", "space", "BackSpace", "Escape"]):
        return False
    return True



__kupfer_name__ = _("Shell Commands")
__kupfer_sources__ = ()
__kupfer_actions__ = (
        "PassToCommand",
        "FilterThroughCommand",
        "WriteToCommand",
    )
__kupfer_text_sources__ = ("CommandTextSource",)
__description__ = _("Run command-line programs. Actions marked with"
                    " the symbol %s run in a subshell.") % "\N{GEAR}"
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os

from kupfer.objects import TextSource, TextLeaf, Action, FileLeaf
from kupfer.objects import OperationError
from kupfer.obj.fileactions import Execute
from kupfer import utils, icons
from kupfer import kupferstring
from kupfer import pretty

def finish_command(ctx, acommand, stdout, stderr, post_result=True):
    """Show async error if @acommand returns error output & error status.
    Else post async result if @post_result.
    """
    max_error_msg=512
    pretty.print_debug(__name__, "Exited:", acommand)
    if acommand.exit_status != 0 and not stdout and stderr:
        errstr = kupferstring.fromlocale(stderr)[:max_error_msg]
        ctx.register_late_error(OperationError(errstr))
    elif post_result:
        leaf = TextLeaf(kupferstring.fromlocale(stdout))
        ctx.register_late_result(leaf)


class GetOutput (Action):
    def __init__(self):
        Action.__init__(self, _("Run (Get Output)"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        if isinstance(leaf, Command):
            argv = ['sh', '-c', leaf.object, '--']
        else:
            argv = [leaf.object]

        def finish_callback(acommand, stdout, stderr):
            finish_command(ctx, acommand, stdout, stderr)

        pretty.print_debug(__name__, "Spawning with timeout 15 seconds")
        utils.AsyncCommand(argv, finish_callback, 15)

    def get_description(self):
        return _("Run program and return its output") + " \N{GEAR}"

class PassToCommand (Action):
    def __init__(self):
        # TRANS: The user starts a program (command) and the text
        # TRANS: is an argument to the command
        Action.__init__(self, _("Pass to Command..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj, ctx):
        self.activate_multiple((leaf,),(iobj, ), ctx)

    def _run_command(self, objs, iobj, ctx):
        if isinstance(iobj, Command):
            argv = ['sh', '-c', iobj.object + ' "$@"', '--']
        else:
            argv = [iobj.object]

        def finish_callback(acommand, stdout, stderr):
            finish_command(ctx, acommand, stdout, stderr, False)

        argv.extend([o.object for o in objs])
        pretty.print_debug(__name__, "Spawning without timeout")
        utils.AsyncCommand(argv, finish_callback, None)

    def activate_multiple(self, objs, iobjs, ctx):
        for iobj in iobjs:
            self._run_command(objs, iobj, ctx)

    def item_types(self):
        yield TextLeaf
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf
        yield Command

    def valid_object(self, iobj, for_item=None):
        if isinstance(iobj, Command):
            return True
        return not iobj.is_dir() and os.access(iobj.object, os.X_OK | os.R_OK)

    def get_description(self):
        return _("Run program with object as an additional parameter") + \
                " \N{GEAR}"


class WriteToCommand (Action):
    def __init__(self):
        # TRANS: The user starts a program (command) and
        # TRANS: the text is written on stdin
        Action.__init__(self, _("Write to Command..."))
        self.post_result = False

    def wants_context(self):
        return True

    def activate(self, leaf, iobj, ctx):
        if isinstance(iobj, Command):
            argv = ['sh', '-c', iobj.object]
        else:
            argv = [iobj.object]

        def finish_callback(acommand, stdout, stderr):
            finish_command(ctx, acommand, stdout, stderr, self.post_result)

        pretty.print_debug(__name__, "Spawning without timeout")
        output = leaf.object.encode("utf-8")
        utils.AsyncCommand(argv, finish_callback, None, stdin=output)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf
        yield Command

    def valid_object(self, iobj, for_item=None):
        if isinstance(iobj, Command):
            return True
        return not iobj.is_dir() and os.access(iobj.object, os.X_OK | os.R_OK)

    def get_description(self):
        return _("Run program and supply text on the standard input") + \
                " \N{GEAR}"

class FilterThroughCommand (WriteToCommand):
    def __init__(self):
        # TRANS: The user starts a program (command) and
        # TRANS: the text is written on stdin, and we
        # TRANS: present the output (stdout) to the user.
        Action.__init__(self, _("Filter through Command..."))
        self.post_result = True

    def get_description(self):
        return _("Run program and supply text on the standard input") + \
                " \N{GEAR}"

class Command (TextLeaf):
    def __init__(self, exepath, name):
        TextLeaf.__init__(self, name, name)
        self.exepath = exepath

    def repr_key(self):
        return self.name

    def get_actions(self):
        yield Execute(quoted=False)
        yield Execute(in_terminal=True, quoted=False)
        yield GetOutput()

    def get_description(self):
        args = " ".join(str(self).split(None, 1)[1:])
        return "%s %s" % (self.exepath, args)

    def get_gicon(self):
        return icons.get_gicon_for_file(self.exepath)

    def get_icon_name(self):
        return "application-x-executable"

class CommandTextSource (TextSource):
    """Yield path and command text items """
    def __init__(self):
        TextSource.__init__(self, name=_("Shell Commands"))

    def get_rank(self):
        return 80

    def get_text_items(self, text):
        if not text.strip():
            return
        if '\n' in text:
            return
        ## check for absolute path with arguments
        firstwords = text.split()
        ## files are handled elsewhere
        if firstwords[0].startswith("/") and len(firstwords) == 1:
            return
        ## absolute paths come out here since
        ## os.path.join with two abspaths returns the latter
        firstword = firstwords[0]
        # iterate over $PATH directories
        PATH = os.environ.get("PATH", os.defpath)
        for execdir in PATH.split(os.pathsep):
            exepath = os.path.join(execdir, firstword)
            if os.access(exepath, os.R_OK|os.X_OK) and os.path.isfile(exepath):
                yield Command(exepath, text)
                break
    def get_description(self):
        return _("Run command-line programs")




def sort_album(album):
    """Sort album in track order"""
    def get_track_number(rec):
        try:
            tnr = int(rec["track-number"])
        except (KeyError, ValueError):
            tnr = 0
        return tnr
    album.sort(key=get_track_number)

def sort_album_order(songs):
    """Sort songs in order by album then by track number

    >>> songs = [
    ... {"title": "a", "album": "B", "track-number": 2},
    ... {"title": "b", "album": "A", "track-number": 1},
    ... {"title": "c", "album": "B", "track-number": 1},
    ... ]
    >>> sort_album_order(songs)
    >>> [s["title"] for s in songs]
    ['b', 'c', 'a']
    """
    def get_album_order(rec):
        return (rec['date'], rec["album"], rec['track-number'])
    songs.sort(key=get_album_order)

def parse_rhythmbox_albums(songs):
    albums = {}
    for song in songs:
        song_artist = song["artist"]
        if not song_artist:
            continue
        song_album = song["album"]
        if not song_album:
            continue
        album = albums.get(song_album, [])
        album.append(song)
        albums[song_album] = album
    # sort album in track order
    for album in albums:
        sort_album(albums[album])
    return albums

def parse_rhythmbox_artists(songs):
    artists = {}
    for song in songs:
        song_artist = song["artist"]
        if not song_artist:
            continue
        artist = artists.get(song_artist, [])
        artist.append(song)
        artists[song_artist] = artist
    # sort in album + track order
    for artist in artists:
        sort_album_order(artists[artist])
    return artists

if __name__ == '__main__':
    import doctest
    doctest.testmod()



import itertools

from collections import OrderedDict

class SavedIterable (object):
    """Wrap an iterable and cache it.

    The SavedIterable can be accessed streamingly, while still being
    incrementally cached. Later attempts to iterate it will access the
    whole of the sequence.

    When it has been cached to its full extent once, it reduces to a
    thin wrapper of a sequence iterator. The SavedIterable will pickle
    into a list.

    >>> s = SavedIterable(range(5))
    >>> next(iter(s))
    0
    >>> list(s)
    [0, 1, 2, 3, 4]

    >>> iter(s)   # doctest: +ELLIPSIS
    <list_iterator object at 0x...>

    >>> import pickle
    >>> pickle.loads(pickle.dumps(s))
    [0, 1, 2, 3, 4]

    >>> u = SavedIterable(list(range(5)))
    >>> one, two = iter(u), iter(u)
    >>> next(one), next(two)
    (0, 0)
    >>> list(two)
    [1, 2, 3, 4]
    >>> list(one)
    [1, 2, 3, 4]

    >>> SavedIterable(list(range(3)))
    [0, 1, 2]
    """
    def __new__(cls, iterable):
        if isinstance(iterable, list):
            return iterable
        return object.__new__(cls)
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.data = []
    def __iter__(self):
        if self.iterator is None:
            return iter(self.data)
        return self._incremental_caching_iter()
    def _incremental_caching_iter(self):
        indices = itertools.count()
        while True:
            idx = next(indices)
            try:
                yield self.data[idx]
            except IndexError:
                pass
            else:
                continue
            if self.iterator is None:
                return
            try:
                x = next(self.iterator)
                self.data.append(x)
                yield x
            except StopIteration:
                self.iterator = None
    def __reduce__(self):
        # pickle into a list with __reduce__
        # (callable, args, state, listitems)
        return (list, (), None, iter(self))

def UniqueIterator(seq, key=None):
    """
    yield items of @seq with set semantics; no duplicates

    >>> list(UniqueIterator([1, 2, 3, 3, 5, 1]))
    [1, 2, 3, 5]
    >>> list(UniqueIterator([1, -2, 3, -3, -5, 2], key=abs))
    [1, -2, 3, -5]
    """
    coll = set()
    if key is None:
        for obj in seq:
            if obj not in coll:
                yield obj
                coll.add(obj)
        return
    else:
        for obj in seq:
            K = key(obj)
            if K not in coll:
                yield obj
                coll.add(K)


class LruCache (object):
    """
    Least-recently-used cache mapping of
    size @maxsiz
    """
    def __init__(self, maxsiz):
        self.d = OrderedDict()
        self.maxsiz = maxsiz

    def __contains__(self, key):
        return key in self.d

    def __setitem__(self, key, value):
        self.d.pop(key, None)
        self.d[key] = value
        if len(self.d) > self.maxsiz:
            # remove the first item (was inserted longest time ago)
            lastkey = next(iter(self.d))
            self.d.pop(lastkey)

    def __getitem__(self, key):
        try:
            value = self.d.pop(key)
        except KeyError:
            raise
        # reinsert the value, puts it "last" in the order
        self.d[key] = value
        return value

if __name__ == '__main__':
    import doctest
    doctest.testmod()




from gi.repository import Gtk, Gdk

from kupfer import version, config


class GetKeyDialogController(object):
    def __init__(self, check_callback=None, previous_key=None,
                 screen=None, parent=None,
                 show_clear=True):
        '''
        check_callback: optional function to check is entered key is valid.
        previous_key - optional previous keybinding, press equal act like cancel
        screen: Screen to use
        parent: Parent toplevel window
        show_clear: Show the “clear” button
        '''
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)

        ui_file = config.get_data_file("getkey_dialog.ui")
        builder.add_from_file(ui_file)
        builder.connect_signals(self)
        self.window = builder.get_object("dialoggetkey")
        self.labelkey = builder.get_object('labelkey')
        self.imagekeybindingaux = builder.get_object('imagekeybindingaux')
        self.labelkeybindingaux = builder.get_object('labelkeybindingaux')
        self.labelaccelerator = builder.get_object('labelaccelerator')
        buttonclear = builder.get_object('buttonclear')
        if not show_clear:
            buttonclear.hide()

        self.imagekeybindingaux.hide()
        self.labelkeybindingaux.hide()

        self._key = None
        self._check_callback = check_callback
        self._previous_key = previous_key
        self._press_time = None

        if screen:
            self.window.set_screen(screen)
        if parent:
            self.window.set_transient_for(parent)
        self.window.connect("focus-in-event", self.on_window_focus_in)
        self.window.connect("focus-out-event", self.on_window_focus_out)

    def run(self):
        ''' Run dialog, return key codes or None when user press cancel'''
        self.window.set_keep_above(True)
        self.window.run()
        self.window.destroy()
        return self._key

    def _return(self, key):
        " Finish dialog with @key as result"
        self._key = key
        self.window.hide()

    def on_buttoncancel_activate(self, _widget):
        self.return_cancel()
        return True

    def on_buttonclear_activate(self, _widget):
        self.return_clear()
        return True

    def return_cancel(self):
        self._return(None)

    def return_clear(self):
        self._return("")

    def translate_keyboard_event(self, widget, event):
        keymap = Gdk.Keymap.get_for_display(widget.get_display())
        # translate keys properly
        _wasmapped, keyval, egroup, level, consumed = keymap.translate_keyboard_state(
                    event.hardware_keycode, event.get_state(), event.group)
        modifiers = Gtk.accelerator_get_default_mod_mask() & ~consumed

        state = event.get_state() & modifiers

        return keyval, state

    def update_accelerator_label(self, keyval, state):
        accel_label = Gtk.accelerator_get_label(keyval, state)
        self.labelaccelerator.set_text(accel_label)

    def on_dialoggetkey_key_press_event(self, widget, event):
        self.imagekeybindingaux.hide()
        self.labelkeybindingaux.hide()
        self._press_time = event.time

        keyval, state = self.translate_keyboard_event(widget, event)
        state = Gdk.ModifierType(state)
        keyname = Gtk.accelerator_name(keyval, state)
        if keyname == 'Escape':
            self.return_cancel()
        elif keyname == 'BackSpace':
            self.return_clear()
        self.update_accelerator_label(keyval, state)

    def on_dialoggetkey_key_release_event(self, widget, event):
        if not self._press_time:
            return
        keyval, state = self.translate_keyboard_event(widget, event)
        self.update_accelerator_label(0, 0)

        state = Gdk.ModifierType(state)
        if Gtk.accelerator_valid(keyval, state):
            key = Gtk.accelerator_name(keyval, state)
            if (self._previous_key is not None and
                    key == self._previous_key):
                self.return_cancel()
                return
            if self._check_callback is None or self._check_callback(key):
                self._return(key)
            else:
                self.imagekeybindingaux.show()
                self.labelkeybindingaux.show()


    def on_window_focus_in(self, window, _event):
        pass

    def on_window_focus_out(self, _window, _event):
        pass


def ask_for_key(check_callback=None, previous_key=None, screen=None,
                parent=None, show_clear=True):
    dlg = GetKeyDialogController(check_callback, previous_key,
                                 screen=screen,
                                 parent=parent,
                                 show_clear=show_clear)
    result = dlg.run()
    return result



__kupfer_name__ = _("GNOME Session Management")
__kupfer_sources__ = ("GnomeItemsSource", )
__description__ = _("Special items and actions for GNOME environment")
__version__ = "2012-10-16"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

"""
Changes:
    2012-10-16 Karol Będkowski:
        + support Gnome3; closes lp#788713;
          author: Joseph Lansdowne
"""

from kupfer.plugin import session_support as support


# sequences of argument lists
LOGOUT_CMD = (["gnome-panel-logout"],
              ["gnome-session-save", "--kill"],
              ["gnome-session-quit", "--logout"])
SHUTDOWN_CMD = (["gnome-panel-logout", "--shutdown"],
                ["gnome-session-save", "--shutdown-dialog"],
                ["gnome-session-quit", "--power-off"])
LOCKSCREEN_CMD = (["gnome-screensaver-command", "--lock"],
                  ["xdg-screensaver", "lock"])

class GnomeItemsSource (support.CommonSource):
    def __init__(self):
        support.CommonSource.__init__(self, _("GNOME Session Management"))
    def get_items(self):
        return (
            support.Logout(LOGOUT_CMD),
            support.LockScreen(LOCKSCREEN_CMD),
            support.Shutdown(SHUTDOWN_CMD),
        )




import os

from gi.repository import Gio 

from kupfer import utils
from kupfer import launch

from kupfer.obj.base import Action, OperationError

class NoDefaultApplicationError (OperationError):
    pass

def is_good_executable(fileleaf):
    if not fileleaf._is_executable():
        return False
    ctype, uncertain = Gio.content_type_guess(fileleaf.object, None)
    return uncertain or Gio.content_type_can_be_executable(ctype)

def get_actions_for_file(fileleaf):
    acts = [GetParent(), ]
    if fileleaf.is_dir():
        acts.append(OpenTerminal())
    elif fileleaf.is_valid():
        if is_good_executable(fileleaf):
            acts.extend((Execute(), Execute(in_terminal=True)))
    return [Open()] + acts

class Open (Action):
    """ Open with default application """
    action_accelerator = "o"
    rank_adjust = 5
    def __init__(self, name=_("Open")):
        Action.__init__(self, name)

    @classmethod
    def default_application_for_leaf(cls, leaf):
        content_attr = Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE
        gfile = leaf.get_gfile()
        info = gfile.query_info(content_attr, Gio.FileQueryInfoFlags.NONE, None)
        content_type = info.get_attribute_string(content_attr)
        def_app = Gio.app_info_get_default_for_type(content_type, False)
        if not def_app:
            raise NoDefaultApplicationError(
                    (_("No default application for %(file)s (%(type)s)") % 
                     {"file": str(leaf), "type": content_type}) + "\n" +
                    _('Please use "%s"') % _("Set Default Application...")
                )
        return def_app

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        self.activate_multiple((leaf, ), ctx)

    def activate_multiple(self, objects, ctx):
        appmap = {}
        leafmap = {}
        for obj in objects:
            app = self.default_application_for_leaf(obj)
            id_ = app.get_id()
            appmap[id_] = app
            leafmap.setdefault(id_, []).append(obj)

        for id_, leaves in leafmap.items():
            app = appmap[id_]
            launch.launch_application(app, paths=[L.object for L in leaves],
                                      activate=False,
                                      screen=ctx and ctx.environment.get_screen())

    def get_description(self):
        return _("Open with default application")

class GetParent (Action):
    action_accelerator = "p"
    rank_adjust = -5
    def __init__(self, name=_("Get Parent Folder")):
        super().__init__(name)
    
    def has_result(self):
        return True

    def activate(self, leaf):
        # Avoid cyclical dep on module level
        from kupfer.objects import FileLeaf
        fileloc = leaf.object
        parent = os.path.normpath(os.path.join(fileloc, os.path.pardir))
        return FileLeaf(parent)

    def get_description(self):
        return None

    def get_icon_name(self):
        return "folder-open"

class OpenTerminal (Action):
    action_accelerator = "t"
    def __init__(self, name=_("Open Terminal Here")):
        super(OpenTerminal, self).__init__(name)

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        try:
            utils.spawn_terminal(leaf.object, ctx.environment.get_screen())
        except utils.SpawnError as exc:
            raise OperationError(exc)

    def get_description(self):
        return _("Open this location in a terminal")
    def get_icon_name(self):
        return "utilities-terminal"

class Execute (Action):
    """ Execute executable file (FileLeaf) """
    rank_adjust = 10
    def __init__(self, in_terminal=False, quoted=True):
        name = _("Run in Terminal") if in_terminal else _("Run (Execute)")
        super(Execute, self).__init__(name)
        self.in_terminal = in_terminal
        self.quoted = quoted

    def repr_key(self):
        return (self.in_terminal, self.quoted)
    
    def activate(self, leaf):
        if self.quoted:
            argv = [leaf.object]
        else:
            argv = utils.argv_for_commandline(leaf.object)
        if self.in_terminal:
            utils.spawn_in_terminal(argv)
        else:
            utils.spawn_async(argv)

    def get_description(self):
        if self.in_terminal:
            return _("Run this program in a Terminal")
        else:
            return _("Run this program")




from kupfer.obj.base import *
from kupfer.obj.exceptions import *
from kupfer.obj.objects import FileLeaf, AppLeaf, UrlLeaf, TextLeaf
from kupfer.obj.objects import RunnableLeaf, SourceLeaf

# Show everything here in help(..)
__all__ = dir()



import functools

import glib
from gi.repository import Gtk

from kupfer import version, config

def idle_call(func):
    @functools.wraps(func)
    def idle_wrapper(*args):
        glib.idle_add(func, *args)
    return idle_wrapper


_HEADER_MARKUP = '<span weight="bold" size="larger">%s</span>'

class ProgressDialogController (object):
    def __init__(self, title, header=None, max_value=100):
        """Create a new progress dialog

        @header: first line of dialog

        The methods show, hide and update are all wrapped to be
        safe to call from any thread.
        """
        self.aborted = False
        self.max_value = float(max_value)
        ui_file = config.get_data_file("progress_dialog.ui")
        self._construct_dialog(ui_file, title, header)

    @idle_call
    def _construct_dialog(self, ui_file, title, header):

        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)

        builder.add_from_file(ui_file)
        builder.connect_signals(self)
        self.window = builder.get_object("window_progress")
        self.button_abort = builder.get_object('button_abort')
        self.progressbar = builder.get_object('progressbar')
        self.label_info = builder.get_object('label_info')
        self.label_header = builder.get_object('label_header')

        self.window.set_title(title)
        if header:
            self.label_header.set_markup(_HEADER_MARKUP %
                    glib.markup_escape_text(header))
        else:
            self.label_header.hide()

        self.update(0, '', '')

    def on_button_abort_clicked(self, widget):
        self.aborted = True
        self.button_abort.set_sensitive(False)

    @idle_call
    def show(self):
        self.window.present()

    @idle_call
    def hide(self):
        self.window.hide()

    @idle_call
    def update(self, value, label, text):
        """ Update dialog information.

        @value: value to set for progress bar
        @label: current action (displayed in emphasized style)
        @text: current information (normal style)
        """
        self.progressbar.set_fraction(min(value/self.max_value, 1.0))
        self.label_info.set_markup("<b>%s</b> %s" %
            (
                glib.markup_escape_text(label),
                glib.markup_escape_text(text),
            ))
        return self.aborted




__kupfer_name__ = _("Instapaper")
__kupfer_sources__ = ()
__kupfer_actions__ = ("SaveToInstapaper",)
__description__ = _("Save url to Instapaper")
__version__ = "2020-11-15"
__author__ = "Peter Stuifzand <peter@p83.nl>"

from kupfer import utils
from kupfer.objects import Action, UrlLeaf, TextLeaf


class SaveToInstapaper(Action):

    def __init__(self):
        Action.__init__(self, _('Save to Instapaper'))

    def activate(self, leaf):
        utils.show_url("https://www.instapaper.com/edit?url=%s" % leaf.object)

    def item_types(self):
        yield UrlLeaf
        yield TextLeaf

    def get_description(self):
        return _("Save url to Instapaper")



"""
session sets up the program as a client to the current
desktop session and enables notifications on session
close, to be able to save state before being killed;

the module API does not depend on the session API used
"""

import os
import time

from gi.repository import GObject
import dbus

from kupfer import pretty, version

class SessionClient (GObject.GObject, pretty.OutputMixin):
    """Session handling controller

    signals:
    save-yourself: Program should save state
    die:           Program should quit immediately
    """
    __gtype_name__ = "SessionClient"

    def __init__(self):
        """Set up program as client to current Session"""
        GObject.GObject.__init__(self)

        succ = False
        try:
            succ = self._connect_session_manager()
        except dbus.DBusException as exc:
            self.output_error(exc)
        if not succ:
            # try to bind to xfce session manager
            try:
                succ = self._connect_xfce_session_manager()
            except dbus.DBusException as exc:
                self.output_error(exc)
        if not succ:
            succ = self._connect_gnomeui()

        # unset autostart id so that it is not transferred
        os.putenv("DESKTOP_AUTOSTART_ID", "")
        self._enabled = succ
        if not self.enabled:
            self.output_info("Warning: Not able to connect to current "
                "desktop session, please Quit before logout to save "
                "kupfer's data.")

    def _connect_gnomeui(self):
        return False

    def _connect_session_manager(self):
        bus = dbus.SessionBus()
        proxy_obj = bus.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus')
        dbus_iface = dbus.Interface(proxy_obj, 'org.freedesktop.DBus')
        service_name = "org.gnome.SessionManager"
        obj_name = "/org/gnome/SessionManager"
        iface_name = service_name

        if not dbus_iface.NameHasOwner(service_name):
            self.output_debug("D-Bus name %s not found" % service_name)
            return False

        try:
            obj = bus.get_object(service_name, obj_name)
        except dbus.DBusException as e:
            pretty.print_error(__name__, e)
            return False
        smanager = dbus.Interface(obj, iface_name)

        app_id = version.PACKAGE_NAME
        startup_id = os.getenv("DESKTOP_AUTOSTART_ID") or ""
        self.client_id = smanager.RegisterClient(app_id, startup_id)
        self._session_ended = False
        self.output_debug("Connected to session as client",
                self.client_id, startup_id)

        private_iface_name = "org.gnome.SessionManager.ClientPrivate"
        bus.add_signal_receiver(self._query_end_session, "QueryEndSession",
                dbus_interface=private_iface_name)
        bus.add_signal_receiver(self._end_session_signal, "EndSession",
                dbus_interface=private_iface_name)
        bus.add_signal_receiver(self._stop_signal, "Stop",
                dbus_interface=private_iface_name)
        return True

    def _connect_xfce_session_manager(self):
        bus = dbus.SessionBus()
        proxy_obj = bus.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus')
        dbus_iface = dbus.Interface(proxy_obj, 'org.freedesktop.DBus')
        service_name = "org.xfce.SessionManager"
        obj_name = "/org/xfce/SessionManager"

        if not dbus_iface.NameHasOwner(service_name):
            self.output_debug("D-Bus name %s not found" % service_name)
            return False

        try:
            bus.get_object(service_name, obj_name)
        except dbus.DBusException as e:
            pretty.print_error(__name__, e)
            return False

        private_iface_name = "org.xfce.Session.Manager"
        bus.add_signal_receiver(self._xfce_session_state_changed, "StateChanged",
                dbus_interface=private_iface_name)
        return True

    def _get_response_obj(self):
        """Return D-Bus session object for ClientPrivate Interface"""
        service_name = "org.gnome.SessionManager"
        obj_name = self.client_id
        iface_name = "org.gnome.SessionManager.ClientPrivate"

        try:
            bus = dbus.Bus()
            obj = bus.get_object(service_name, obj_name)
        except dbus.DBusException as e:
            pretty.print_error(__name__, e)
            return None
        smanager = dbus.Interface(obj, iface_name)
        return smanager

    def _query_end_session(self, flags):
        self.output_debug("Query end", flags)

        smanager = self._get_response_obj()
        smanager and smanager.EndSessionResponse(True, "Always OK")

    def _end_session_signal(self, flags):
        self.output_debug("Session end", flags)
        if not self._session_ended:
            self._session_ended = True
            self.emit("save-yourself")
        smanager = self._get_response_obj()
        smanager and smanager.EndSessionResponse(True, "Always OK")

    def _xfce_session_state_changed(self, old_value, new_value):
        self.output_debug("XFCE Session change", time.asctime(),
                old_value, new_value)
        if new_value == 4:
            self.emit("save-yourself")

    def _stop_signal(self):
        self.output_debug("Session stop")
        self.emit("die")
    def _session_save(self, obj, *args):
        self.emit("save-yourself")
    def _session_die(self, obj, *args):
        self.emit("die")
    @property
    def enabled(self):
        """If a session connection is available"""
        return self._enabled

GObject.signal_new("save-yourself", SessionClient, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())
GObject.signal_new("die", SessionClient, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())



__kupfer_name__ = _("Firefox Bookmarks")
__kupfer_sources__ = ("BookmarksSource", )
__kupfer_actions__ = ()
__description__ = _("Index of Firefox bookmarks")
__version__ = "2020.1"
__author__ = "Ulrik, William Friesen, Karol Będkowski"

from contextlib import closing
import os
import sqlite3
import time

from kupfer.objects import Source
from kupfer.objects import UrlLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer import plugin_support

from ._firefox_support import get_firefox_home_file

MAX_ITEMS = 10000

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "profile",
        "label": _("Firefox profile name or path"),
        "type": str,
        "value": "",
    },
)


class BookmarksSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("firefox", "firefox-esr")
    def __init__(self):
        super().__init__(_("Firefox Bookmarks"))
        self._version = 3

    def initialize(self):
        ff_home = get_firefox_home_file('')
        self.monitor_token = self.monitor_directories(ff_home)

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == 'lock'

    def _get_ffx3_bookmarks(self):
        """Query the firefox places bookmark database"""
        profile = __kupfer_settings__["profile"]
        fpath = get_firefox_home_file("places.sqlite", profile)
        if not (fpath and os.path.isfile(fpath)):
            return []

        fpath = fpath.replace("?", "%3f").replace("#", "%23")
        fpath = "file:" + fpath + "?immutable=1&mode=ro"

        for _ in range(2):
            try:
                self.output_debug("Reading bookmarks from", fpath)
                with closing(sqlite3.connect(fpath, uri=True,
                                             timeout=1)) as conn:
                    c = conn.cursor()
                    c.execute("""SELECT moz_places.url, moz_bookmarks.title
                              FROM moz_places, moz_bookmarks
                              WHERE moz_places.id = moz_bookmarks.fk
                                    AND moz_bookmarks.keyword_id IS NULL
                              ORDER BY visit_count DESC
                              LIMIT ?""",
                              (MAX_ITEMS, ))
                    return [UrlLeaf(url, title) for url, title in c]
            except sqlite3.Error as err:
                # Something is wrong with the database
                # wait short time and try again
                self.output_debug("Read bookmarks error:", str(err))
                time.sleep(1)
        self.output_exc()
        return []

    def get_items(self):
        return self._get_ffx3_bookmarks()

    def get_description(self):
        return _("Index of Firefox bookmarks")

    def get_gicon(self):
        return self.get_leaf_repr() and self.get_leaf_repr().get_gicon()

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield UrlLeaf



__kupfer_name__ = _("Firefox Tags")
__kupfer_sources__ = ("TagsSource",)
__description__ = _("Browse Firefox bookmarks by tags")
__version__ = "2021-09-05"
__author__ = "Karol Będkowski"

from contextlib import closing
import os
import sqlite3
import time

from kupfer import plugin_support
from kupfer.objects import Source, Leaf, UrlLeaf
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.obj.apps import AppLeafContentMixin

from ._firefox_support import get_firefox_home_file

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "profile",
        "label": _("Firefox profile name or path"),
        "type": str,
        "value": "",
    },
)


MAX_ITEMS = 10000


class FirefoxTag(Leaf):
    def get_icon_name(self):
        return "text-html"

    def has_content(self):
        return True

    def content_source(self, alternate=False):
        return BookmarksSource(self.object, self.name)


class TagsSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("firefox", "firefox-esr")

    def __init__(self):
        super().__init__(_("Firefox Tags"))
        self.monitor_token = None

    def initialize(self):
        profile = __kupfer_settings__["profile"]
        ff_home = get_firefox_home_file("", profile)
        self.monitor_token = self.monitor_directories(ff_home)

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == "lock"

    def get_items(self):
        """Get tags from firefox places database"""
        profile = __kupfer_settings__["profile"]
        fpath = get_firefox_home_file("places.sqlite", profile)
        if not (fpath and os.path.isfile(fpath)):
            return []

        fpath = fpath.replace("?", "%3f").replace("#", "%23")
        fpath = "file:" + fpath + "?immutable=1&mode=ro"

        for _ in range(2):
            try:
                self.output_debug("Reading bookmarks from", fpath)
                with closing(
                    sqlite3.connect(fpath, uri=True, timeout=1)
                ) as conn:
                    cur = conn.cursor()
                    cur.execute(
                        "SELECT id, title FROM moz_bookmarks mb "
                        "WHERE mb.parent = 4 AND mb.fk IS NULL"
                    )
                    return [FirefoxTag(tagid, title) for tagid, title in cur]
            except sqlite3.Error as err:
                self.output_debug("Read bookmarks error:", str(err))
                # Something is wrong with the database
                # wait short time and try again
                time.sleep(1)

        self.output_exc()
        return []

    def get_description(self):
        return _("Index of Firefox bookmarks by tags")

    def get_gicon(self):
        return self.get_leaf_repr() and self.get_leaf_repr().get_gicon()

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield FirefoxTag


class BookmarksSource(Source):
    def __init__(self, tag_id, tag):
        super().__init__(_("Firefox Bookmarks by tag"))
        self.tag = tag
        self.tag_id = tag_id

    def get_items(self):
        """Query the firefox places database for bookmarks with tag"""

        profile = __kupfer_settings__["profile"]
        fpath = get_firefox_home_file("places.sqlite", profile)
        if not (fpath and os.path.isfile(fpath)):
            return []

        fpath = fpath.replace("?", "%3f").replace("#", "%23")
        fpath = "file:" + fpath + "?immutable=1&mode=ro"

        for _ in range(2):
            try:
                self.output_debug("Reading bookmarks from", fpath)
                with closing(
                    sqlite3.connect(fpath, uri=True, timeout=1)
                ) as conn:
                    cur = conn.cursor()
                    cur.execute(
                        """
SELECT mp.url, mp.title
FROM moz_places mp
JOIN moz_bookmarks mb ON mp.id = mb.fk
WHERE mb.keyword_id IS NULL
	AND	mb.parent = ?
ORDER BY visit_count DESC
LIMIT ?""",
                        (self.tag_id, MAX_ITEMS),
                    )
                    return [UrlLeaf(url, title) for url, title in cur]
            except sqlite3.Error as err:
                # Something is wrong with the database
                # wait short time and try again
                self.output_debug("Read bookmarks error:", str(err))
                time.sleep(1)

        self.output_exc()
        return []

    def get_gicon(self):
        return self.get_leaf_repr() and self.get_leaf_repr().get_gicon()

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield UrlLeaf



from gi.repository import Gtk, Gdk

class TextRepresentation (object):
    """
    Kupfer Objects that implement this interface have a plain text
    representation that can be used for Copy & Paste etc
    """
    def get_text_representation(self):
        """The default implementation returns the represented object"""
        return self.object

class UriListRepresentation (object):
    """
    Kupfer Objects that implement this interface have a uri-list
    representation that can be used for Copy & Paste etc

    get_urilist_representation should return a sequence of bytestring
    URIs.
    """
    def get_urilist_representation(self):
        """The default implementation raises notimplementederror """
        raise NotImplementedError

def get_text_representation(obj):
    try:
        return obj.get_text_representation()
    except AttributeError:
        return None

def copy_to_clipboard(obj, clipboard):
    """
    Copy @obj to @clipboard, a Gtk.Clipboard

    Return True if successful
    """
    ## support copying text to clipboard
    ## as well as files in both the general uri-list representation
    ## and in nautilus' file copy clipboard type
    (uri_id, text_id, nautilus_id) = (80, 81, 82)
    nautilus_target = Gdk.Atom.intern('x-special/gnome-copied-files', False)

    # udata is the data dict
    def store(clipboard, sdata, info, udata):
        if info == uri_id:
            sdata.set_uris(udata[uri_id])
        if info == text_id:
            sdata.set_text(udata[text_id])
        if info == nautilus_id:
            str_data_format = 8
            sdata.set(nautilus_target, str_data_format, udata[nautilus_id])
    def clear(clipboard, udata):
        pass

    targets = []
    data = {}
    try:
        urilist = obj.get_urilist_representation()
    except AttributeError:
        pass
    else:
        if urilist:
            targets = Gtk.TargetList.new(None)
            targets.add_uri_targets(uri_id)
            #targets = Gtk.target_list_add_uri_targets(targets, uri_id)
            targets.add(nautilus_target, 0, nautilus_id)
            data[uri_id] = urilist
            data[nautilus_id] = 'copy\n' + '\n'.join(urilist)

    try:
        text = obj.get_text_representation()
    except AttributeError:
        pass
    else:
        targets = Gtk.TargetList.new(None)
        targets.add_text_targets(text_id)
        clipboard.set_text(text, -1) # -1 for computed string length
        data[text_id] = text
    if data:
        # FIXME: How to set URIs on clipboard?
        #clipboard.set_with_data(targets, store, clear, data)
        # store all targets
        #clipboard.set_can_store(targets)
        return True
    return False

def get_fileleaf_for_path(pth):
    import kupfer.objects
    return kupfer.objects.FileLeaf(pth)



__kupfer_name__ = _("Thunar")
__kupfer_sources__ = ("ThunarObjects", )
__kupfer_actions__ = (
    "Reveal",
    "GetInfo",
    "SendTo",
    "CopyTo",
    "LinkTo",
    "MoveTo",
)
__description__ = _("File manager Thunar actions")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os

import dbus
from gi.repository import Gio

from kupfer.objects import Action, Source
from kupfer.objects import InvalidDataError, NotAvailableError, NoMultiError
from kupfer.objects import FileLeaf, RunnableLeaf, AppLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer import config
from kupfer import plugin_support
from kupfer import pretty

plugin_support.check_dbus_connection()

SERVICE_NAME = "org.xfce.Thunar"
OBJECT_PATH = "/org/xfce/FileManager"
IFACE_NAME = "org.xfce.FileManager"

TRASH_IFACE_NAME = "org.xfce.Trash"

def _get_thunar():
    """Return the dbus proxy object for Thunar
    we will activate it over d-bus (start if not running)
    """
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(SERVICE_NAME, OBJECT_PATH)
    except dbus.DBusException as e:
        pretty.print_error(__name__, e)
        return
    iface_obj = dbus.Interface(proxy_obj, IFACE_NAME)
    return iface_obj

def _get_thunar_trash():
    """Return the dbus proxy object for Thunar
    we will activate it over d-bus (start if not running)
    """
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(SERVICE_NAME, OBJECT_PATH)
    except dbus.DBusException as e:
        pretty.print_error(__name__, e)
        return
    iface_obj = dbus.Interface(proxy_obj, TRASH_IFACE_NAME)
    return iface_obj

def _dummy(*args):
    pass

class Reveal (Action):
    def __init__(self):
        Action.__init__(self, _("Select in File Manager"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        gfile = leaf.get_gfile()
        parent = gfile.get_parent()
        if not parent:
            return
        uri = parent.get_uri()
        bname = gfile.get_basename()
        id_ = ctx.environment.get_startup_notification_id()
        display = ctx.environment.get_display()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            _get_thunar().DisplayFolderAndSelect(uri, bname, display,
                id_, reply_handler=_dummy, error_handler=_dummy)
        except TypeError:
            # Thunar 1.0 Uses $DISPLAY
            _get_thunar().DisplayFolderAndSelect(uri, bname, display,
                reply_handler=_dummy, error_handler=_dummy)

    def item_types(self):
        yield FileLeaf

class GetInfo (Action):
    def __init__(self):
        Action.__init__(self, _("Show Properties"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        gfile = leaf.get_gfile()
        uri = gfile.get_uri()
        id_ = ctx.environment.get_startup_notification_id()
        display = ctx.environment.get_display()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            _get_thunar().DisplayFileProperties(uri, display,
                id_, reply_handler=_dummy, error_handler=_dummy)
        except TypeError:
            # Thunar 1.0 Uses $DISPLAY
            _get_thunar().DisplayFileProperties(uri, display,
                reply_handler=_dummy, error_handler=_dummy)

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Show information about file in file manager")

    def get_icon_name(self):
        return "dialog-information"


class SendTo (Action):
    """ Send files to  selected app from "send to" list """
    def __init__(self):
        Action.__init__(self, _("Send To..."))

    def activate_multiple(self, leaves, iobjs):
        for app in iobjs:
            app.launch(paths=[leaf.object for leaf in leaves])

    def activate(self, leaf, iobj):
        self.activate_multiple((leaf, ), (iobj, ))

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, for_item=None):
        return _SendToAppsSource()

def _good_destination(dpath, spath):
    """If directory path @dpath is a valid destination for file @spath
    to be copied or moved to. 
    """
    if not os.path.isdir(dpath):
        return False
    spath = os.path.normpath(spath)
    dpath = os.path.normpath(dpath)
    cpfx = os.path.commonprefix((spath, dpath))
    if os.path.samefile(dpath, spath) or cpfx == spath:
        return False
    return True

def leaf_uri(leaf):
    return leaf.get_gfile().get_uri()

class CopyTo (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Copy To..."))

    def wants_context(self):
        return True

    def activate_multiple(self, leaves, iobjects, ctx):
        # Unroll by looping over the destinations,
        # copying everything into each destination
        thunar = _get_thunar()
        work_dir = os.path.expanduser("~/")
        display = ctx.environment.get_display()
        notify_id = ctx.environment.get_startup_notification_id()
        sourcefiles = [leaf_uri(L) for L in leaves]

        def _reply(*args):
            self.output_debug("reply got for copying", *args)

        def _reply_error(exc):
            self.output_debug(exc)
            ctx.register_late_error(NotAvailableError(_("Thunar")))

        for dest_iobj in iobjects:
            desturi = leaf_uri(dest_iobj)
            thunar.CopyInto(work_dir, sourcefiles, desturi, display, notify_id,
                            reply_handler=_reply,
                            error_handler=_reply_error)

    def activate(self, leaf, iobj, ctx):
        return self.activate_multiple([leaf], [iobj], ctx)

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return True
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)
    def get_description(self):
        return _("Copy file to a chosen location")

class MoveTo (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Move To..."))

    def wants_context(self):
        return True

    def activate_multiple(self, leaves, iobjects, ctx):
        if len(iobjects) != 1:
            raise NoMultiError()

        def _reply():
            self.output_debug("reply got for moving")

        def _reply_error(exc):
            self.output_debug(exc)
            ctx.register_late_error(NotAvailableError(_("Thunar")))

        (dest_iobj,) = iobjects
        # Move everything into the destination
        thunar = _get_thunar()
        work_dir = os.path.expanduser("~/")
        display = ctx.environment.get_display()
        notify_id = ctx.environment.get_startup_notification_id()
        sourcefiles = [leaf_uri(L) for L in leaves]
        desturi = leaf_uri(dest_iobj)
        thunar.MoveInto(work_dir, sourcefiles, desturi, display, notify_id,
                        reply_handler=_reply,
                        error_handler=_reply_error)

    def activate(self, leaf, iobj, ctx):
        return self.activate_multiple([leaf], [iobj], ctx)

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return True
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)
    def get_description(self):
        return _("Move file to new location")
    def get_icon_name(self):
        return "go-next"

class LinkTo (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Symlink In..."))

    def wants_context(self):
        return True

    def activate_multiple(self, leaves, iobjects, ctx):
        # Unroll by looping over the destinations,
        # copying everything into each destination
        thunar = _get_thunar()
        work_dir = os.path.expanduser("~/")
        display = ctx.environment.get_display()
        notify_id = ctx.environment.get_startup_notification_id()
        sourcefiles = [leaf_uri(L) for L in leaves]

        def _reply(*args):
            self.output_debug("reply got for copying", *args)

        def _reply_error(exc):
            self.output_debug(exc)
            ctx.register_late_error(NotAvailableError(_("Thunar")))

        for dest_iobj in iobjects:
            desturi = leaf_uri(dest_iobj)
            thunar.LinkInto(work_dir, sourcefiles, desturi, display, notify_id,
                            reply_handler=_reply,
                            error_handler=_reply_error)

    def activate(self, leaf, iobj, ctx):
        return self.activate_multiple([leaf], [iobj], ctx)

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return True
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)
    def get_description(self):
        return _("Create a symlink to file in a chosen location")

class EmptyTrash (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Empty Trash"))

    def wants_context(self):
        return True

    def run(self, ctx):
        id_ = ctx.environment.get_startup_notification_id()
        thunar = _get_thunar_trash()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            thunar.EmptyTrash(ctx.environment.get_display(), id_,
                reply_handler=_dummy, error_handler=_dummy)
        except TypeError:
            # Thunar 1.0 uses only $DISPLAY arg
            thunar.EmptyTrash(ctx.environment.get_display(),
                reply_handler=_dummy, error_handler=_dummy)

    def get_description(self):
        return None
    def get_icon_name(self):
        return "user-trash-full"

class ThunarObjects (AppLeafContentMixin, Source):
    appleaf_content_id = "Thunar"
    def __init__(self):
        Source.__init__(self, _("Thunar"))

    def get_items(self):
        yield EmptyTrash()

    def provides(self):
        yield RunnableLeaf

    def get_icon_name(self):
        return "Thunar"


class _SendToAppsSource (Source):
    """ Send To items source """
    def __init__(self):
        Source.__init__(self, _("Thunar Send To Objects"))

    def get_items(self):
        for data_dir in config.get_data_dirs("sendto", package="Thunar"):
            for filename in os.listdir(data_dir):
                if not filename.endswith('.desktop'):
                    continue
                file_path = os.path.join(data_dir, filename)
                if not os.path.isfile(file_path):
                    continue
                try:
                    yield AppLeaf(init_path=file_path, require_x=False)
                except InvalidDataError:
                    pass

    def get_icon_name(self):
        return "Thunar"

    def provides(self):
        yield AppLeaf



__kupfer_name__ = _("Chromium Bookmarks")
__kupfer_sources__ = ("BookmarksSource", )
__description__ = _("Index of Chromium bookmarks")
__version__ = ""
__author__ = "Francesco Marella <francesco.marella@gmail.com>"

from kupfer.objects import Source
from kupfer.objects import UrlLeaf
from kupfer import config
from kupfer.obj.apps import AppLeafContentMixin


class BookmarksSource (AppLeafContentMixin, Source):
    appleaf_content_id = ("chromium-browser", "chromium")
    def __init__(self):
        super(BookmarksSource, self).__init__(_("Chromium Bookmarks"))

    def _get_chromium_items(self, fpath):
        """Parse Chromium' bookmarks backups"""
        from kupfer.plugin import chromium_support
        self.output_debug("Parsing", fpath)
        bookmarks = chromium_support.get_bookmarks(fpath)
        for book in bookmarks:
            yield UrlLeaf(book["url"], book["name"])

    def get_items(self):
        fpath = config.get_config_file("Bookmarks", package="chromium/Default")

        # If there is no chromium bookmarks file, look for a google-chrome one
        if not fpath:
            fpath = config.get_config_file("Bookmarks",package="google-chrome/Default")

        if fpath:
            try:
                return self._get_chromium_items(fpath)
            except Exception as exc:
                self.output_error(exc)

        self.output_error("No Chromium bookmarks file found")
        return []

    def get_description(self):
        return _("Index of Chromium bookmarks")
    def get_icon_name(self):
        return "chromium-browser"
    def provides(self):
        yield UrlLeaf



"""
This is a simple plugin demonstration, how to add single, simple actions
"""

__kupfer_name__ = _("Wikipedia")
__kupfer_sources__ = ()
__kupfer_actions__ = ("WikipediaSearch", )
__description__ = _("Search in Wikipedia")
__version__ = "2017.1"
__author__ = "US"

import urllib.parse

from kupfer.objects import Action, TextLeaf
from kupfer import utils, plugin_support


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "lang",
        "label": _("Wikipedia language"),
        "type": str,
        # TRANS: Default wikipedia language code
        "value": _("en"),
    },
)


class WikipediaSearch (Action):
    def __init__(self):
        Action.__init__(self, _("Search in Wikipedia"))

    def activate(self, leaf):
        # Send in UTF-8 encoding
        lang_code = __kupfer_settings__["lang"]
        search_url="https://%s.wikipedia.org/w/index.php?title=Special:Search&go=Go" % lang_code
        # will encode search=text, where `text` is escaped
        query_url = search_url + "&" + urllib.parse.urlencode({"search": leaf.object})
        utils.show_url(query_url)
    def item_types(self):
        yield TextLeaf
    def get_description(self):
        lang_code = __kupfer_settings__["lang"]
        return _("Search for this term in %s.wikipedia.org") % lang_code
    def get_icon_name(self):
        return "edit-find"




"""Firefox common functions."""

import os
from configparser import RawConfigParser

from kupfer import pretty


def make_absolute_and_check(firefox_dir, path):
    """Helper, make path absolute and check is exist."""
    if not path.startswith("/"):
        path = os.path.join(firefox_dir, path)

    if os.path.isdir(path):
        return path

    return None


def _find_default_profile(firefox_dir):
    """Try to find default/useful profile in firefox located in `firefox_dir`
    """
    config = RawConfigParser({"Default" : 0})
    config.read(os.path.join(firefox_dir, "profiles.ini"))
    path = None

    # find Instal.* section and default profile
    for section in config.sections():
        if section.startswith("Install"):
            if not config.has_option(section, "Default"):
                continue

            # found default profile
            path = make_absolute_and_check(firefox_dir,
                                           config.get(section, "Default"))
            if path:
                pretty.print_debug(__name__, "found install default profile",
                                   path)
                return path

            break

    pretty.print_debug("Install* default profile not found")

    # not found default profile, iterate profiles, try to find default
    for section in config.sections():
        if not section.startswith("Profile"):
            continue

        if config.has_option(section, "Default") and \
                config.get(section, "Default") == "1":
            path = make_absolute_and_check(firefox_dir,
                                           config.get(section, "Path"))
            if path:
                pretty.print_debug(__name__, "Found profile with default=1",
                                   section, path)
                break

        if not path and config.has_option(section, "Path"):
            path = make_absolute_and_check(firefox_dir,
                                           config.get(section, "Path"))

    return path


def get_firefox_home_file(needed_file, profile_dir=None):
    """Get path to `needed_file` in `profile_dir`.

        When no `profile_dir` is not given try to find default profile
        in profiles.ini. `profile_dir` may be only profile name and is relative
        to ~/.mozilla/firefox or may be full path to profile dir.
    """
    if profile_dir:
        profile_dir = os.path.expanduser(profile_dir)
        if not os.path.isabs(profile_dir):
            profile_dir = os.path.join(
                os.path.expanduser("~/.mozilla/firefox"), profile_dir)

        if not os.path.isdir(profile_dir):
            pretty.print_debug(__name__,
                               "Firefox custom profile_dir not exists",
                               profile_dir)
            return None

        return os.path.join(profile_dir, needed_file)

    firefox_dir = os.path.expanduser("~/.mozilla/firefox")
    if not os.path.exists(firefox_dir):
        pretty.print_debug(__name__, "Firefox dir not exists", firefox_dir)
        return None

    if not os.path.isfile(os.path.join(firefox_dir, "profiles.ini")):
        pretty.print_debug(__name__, "Firefox profiles.ini not exists",
                           firefox_dir)
        return None

    pretty.print_debug(__name__, "Firefox dir", firefox_dir)

    path = _find_default_profile(firefox_dir)
    pretty.print_debug(__name__, "Profile path", path)

    return os.path.join(path, needed_file) if path else None



__kupfer_name__ = _("Kupfer Plugins")
__kupfer_sources__ = ("KupferPlugins", )
__description__ = _("Access Kupfer's plugin list in Kupfer")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os

from kupfer.objects import Action, Source, Leaf, FileLeaf, TextLeaf
from kupfer import kupferui

# Since this is a core plugin we break some rules
# These modules are normally out of bounds for plugins
from kupfer.core import plugins, settings


class ShowInfo (Action):
    def __init__(self):
        Action.__init__(self, _("Show Information"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        plugin_id = leaf.object["name"]
        kupferui.show_plugin_info(plugin_id, ctx.environment)

    def get_description(self):
        pass
    def get_icon_name(self):
        return "dialog-information"

class ShowSource (Action):
    def __init__(self):
        Action.__init__(self, _("Show Source Code"))

    def has_result(self):
        return True
    def activate(self, leaf):
        # Try to find the __file__ attribute for the plugin
        # It will fail for files inside zip packages, but that is
        # uncommon for now.
        # Additionally, it will fail for fake plugins
        plugin_id = leaf.object["name"]
        filename = plugins.get_plugin_attribute(plugin_id, "__file__")
        if not filename:
            return leaf
        root, ext = os.path.splitext(filename)
        if ext.lower() == ".pyc" and os.path.exists(root + ".py"):
            return FileLeaf(root + ".py")

        if not os.path.exists(filename):
            # handle modules in zip or eggs
            import pkgutil
            pfull = "kupfer.plugin." + plugin_id
            loader = pkgutil.get_loader(pfull)
            if loader:
                return TextLeaf(loader.get_source(pfull))
        return FileLeaf(filename)

    def get_description(self):
        pass
    def get_icon_name(self):
        return "dialog-information"

class Plugin (Leaf):
    # NOTE: Just to be sure that a plugin ranks lower than a
    # like-named other object by default.
    rank_adjust = -1
    def __init__(self, obj, name):
        Leaf.__init__(self, obj, name)
    def get_actions(self):
        yield ShowInfo()
        yield ShowSource()

    def get_description(self):
        setctl = settings.GetSettingsController()
        enabled = setctl.get_plugin_enabled(self.object["name"])
        return "%s (%s)" % (self.object["description"],
                _("enabled") if enabled else _("disabled"))
    def get_icon_name(self):
        return "package-x-generic"

class KupferPlugins (Source):
    def __init__(self):
        Source.__init__(self, _("Kupfer Plugins"))

    def get_items(self):
        setctl = settings.GetSettingsController()
        for info in plugins.get_plugin_info():
            plugin_id = info["name"]
            if setctl.get_plugin_is_hidden(plugin_id):
                continue
            yield Plugin(info, info["localized_name"])

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield Plugin

    def get_icon_name(self):
        return "package-x-generic"

    @classmethod
    def decorates_type(cls):
        return Plugin

    @classmethod
    def decorate_item(cls, obj):
        if cls.is_self_plugin(obj):
            return cls()

    @classmethod
    def is_self_plugin(cls, obj):
        self_plug_id = __name__.split(".")[-1]
        return obj.object['name'] == self_plug_id

    def get_leaf_repr(self):
        for obj in self.get_leaves():
            if self.is_self_plugin(obj):
                return obj




import os
import urllib.parse
import urllib.request, urllib.parse, urllib.error

from kupfer.objects import TextSource, TextLeaf, FileLeaf, UrlLeaf
from kupfer.obj.objects import OpenUrl
from kupfer import utils

__kupfer_name__ = "Free-text Queries"
__kupfer_sources__ = ()
__kupfer_text_sources__ = ("BasicTextSource", "PathTextSource", "URLTextSource",)
__kupfer_actions__ = ("OpenTextUrl", )
__description__ = "Basic support for free-text queries"
__version__ = "2021.1"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

class BasicTextSource (TextSource):
    """The most basic TextSource yields one TextLeaf"""
    def __init__(self):
        TextSource.__init__(self, name=_("Text"))

    def get_text_items(self, text):
        if not text:
            return
        yield TextLeaf(text)
    def provides(self):
        yield TextLeaf


class PathTextSource (TextSource):
    """Return existing full paths if typed"""
    def __init__(self):
        TextSource.__init__(self, name="Filesystem Text Matches")

    def get_rank(self):
        return 80

    def _get_hostname(self):
        if not hasattr(self, '_hostname'):
            try:
                import socket
                self._hostname = socket.gethostname()
            except Exception:
                self._hostname = ""
                self.output_exc()
        return self._hostname

    def _is_local_file_url(self, url):
        # Recognize file:/// or file://localhost/ or file://<local_hostname>/ URLs
        if url.startswith("file:"):
            hostname = self._get_hostname()
            for prefix in ['file:///', 'file://localhost/', 'file://%s/' % hostname]:
                if url.startswith(prefix):
                    return True
        return False

    def get_text_items(self, text):
        # Find directories or files
        if self._is_local_file_url(text):
            leaf = FileLeaf.from_uri(text)
            if leaf and leaf.is_valid():
                yield leaf
        else:
            prefix = os.path.expanduser("~/")
            ufilepath = text if os.path.isabs(text) else os.path.join(prefix, text)
            filepath = os.path.normpath(ufilepath)
            if os.access(filepath, os.R_OK):
                yield FileLeaf(filepath)

    def provides(self):
        yield FileLeaf

def is_url(text):
    """If @text is an URL, return a cleaned-up URL, else return None"""
    text = text.strip()
    components = list(urllib.parse.urlparse(text))
    domain = "".join(components[1:])
    dotparts = domain.rsplit(".")

    # 1. Domain name part is one word (without spaces)
    # 2. Urlparse parses a scheme (http://), else we apply heuristics
    if len(domain.split()) == 1 and (components[0] or ("." in domain and
        len(dotparts) >= 2 and len(dotparts[-1]) >= 2 and
        any(char.isalpha() for char in domain) and
        all(part[:1].isalnum() for part in dotparts))):
        if not components[0]:
            url = "http://" + "".join(components[1:])
        else:
            url = text
        name = ("".join(components[1:3])).strip("/")
        if name:
            return url

def try_unquote_url(url):
    """Try to turn an URL-escaped string into a Unicode string

    Where we assume UTF-8 encoding; and return the original url if
    any step fails.
    """
    return urllib.parse.unquote(url)

class OpenTextUrl (OpenUrl):
    rank_adjust = 1

    def activate(self, leaf):
        url = is_url(leaf.object)
        utils.show_url(url)

    def item_types(self):
        yield TextLeaf
    def valid_for_item(self, leaf):
        return is_url(leaf.object)

class URLTextSource (TextSource):
    """detect URLs and webpages"""
    def __init__(self):
        TextSource.__init__(self, name="URL Text Matches")

    def get_rank(self):
        return 75
    def get_text_items(self, text):
        # Only detect "perfect" URLs
        text = text.strip()
        components = list(urllib.parse.urlparse(text))
        domain = "".join(components[1:])

        # If urlparse parses a scheme (http://), it's an URL
        if len(domain.split()) <= 1 and components[0]:
            url = text
            name = ("".join(components[1:3])).strip("/")
            name = try_unquote_url(name) or url
            yield UrlLeaf(url, name=name)

    def provides(self):
        yield UrlLeaf



# -*- encoding: UTF-8 -*-
"""
Classes used to provide grouping leaves mechanism.
"""
import copy
import itertools
import time
import weakref

from kupfer.objects import Leaf, Source
from kupfer import utils

__author__ = ("Karol Będkowski <karol.bedkowsk+gh@gmail.com>, "
              "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>" )

class GroupingLeaf (Leaf):
    """
    A Leaf that groups with other leaves inside Grouping Sources

    The represented object of a GroupedLeaf is a dictionary of (slot, value)
    pairs, where slot identifies the slot, and the value is something that must
    be equal to be grouped.

    The GroupingLeaf must have a value for all @grouping_slots, but values of
    None will not be grouped with others.
    """
    grouping_slots = ()

    def __init__(self, obj, name):
        Leaf.__init__(self, obj, name)
        self.links = [self]

    def slots(self):
        return self.object

    def has_content(self):
        return len(self.links) > 1

    def content_source(self, alternate=False):
        return _GroupedItemsSource(self)

    def __len__(self):
        return len(self.links)

    def __contains__(self, key):
        "Return True if GroupedLeaf has value for @key"
        return any(key in leaf.object for leaf in self.links)

    def __getitem__(self, key):
        "Get first (canonical) value for key"
        try:
            return next(iter(self.all(key)))
        except StopIteration:
            raise KeyError("%s has no slot %s" % (self, key))

    def all(self, key):
        "Return iterator of all values for @key"
        return (leaf.object[key] for leaf in self.links if key in leaf.object)

    def check_key(self, key):
        ''' check if GroupedLeaf has non empty value for @key '''
        return any(bool(leaf.object.get(key)) for leaf in self.links)

class GroupingSource (Source):

    def __init__(self, name, sources):
        Source.__init__(self, name)
        self.sources = sources

    def get_leaves(self, force_update=False):
        starttime = time.time()
        # map (slot, value) -> group
        groups = {}
        non_group_leaves = []
        for src in self.sources:
            leaves = Source.get_leaves(src, force_update)
            for leaf in leaves:
                try:
                    slots = leaf.slots()
                except AttributeError:
                    # Let through Non-grouping leaves
                    non_group_leaves.append(leaf)
                    continue
                slots = leaf.slots()
                for slot in leaf.grouping_slots:
                    value = slots.get(slot)
                    if value:
                        groups.setdefault((slot, value), set()).add(leaf)
                if not leaf.grouping_slots:
                    self.output_error("GroupingLeaf has no grouping slots",
                            repr(leaf))

        # Keep track of keys that are only duplicate references
        redundant_keys = set()

        def merge_groups(key1, key2):
            if groups[key1] is groups[key2]:
                return
            groups[key1].update(groups[key2])
            groups[key2] = groups[key1]
            redundant_keys.add(key2)

        # Find all (slot, value) combinations that have more than one leaf
        # and merge those groups
        for (slot, value), leaves in groups.items():
            if len(leaves) <= 1:
                continue
            for leaf in list(leaves):
                for slot2 in leaf.grouping_slots:
                    for value2 in leaf.all(slot2):
                        if not value2:
                            continue
                        merge_groups((slot, value), (slot2, value2))
        if self.should_sort_lexically():
            sort_func = utils.locale_sort
        else:
            sort_func = lambda x: x

        keys = set(groups)
        keys.difference_update(redundant_keys)
        leaves = sort_func(self._make_group_leader(groups[K]) for K in keys)
        mergetime = time.time() - starttime
        if mergetime > 0.05:
            self.output_debug("Warning(?): merged in %s seconds" % mergetime)
        return itertools.chain(non_group_leaves, leaves)

    def repr_key(self):
        # Distinguish when used as GroupingSource
        if type(self) is GroupingSource:
            return str(self)
        return Source.repr_key(self)

    @classmethod
    def _make_group_leader(cls, leaves):
        if len(leaves) == 1:
            (leaf, ) = leaves
            return leaf
        obj = copy.copy(next(iter(leaves)))
        obj.links = list(leaves)
        for other in leaves:
            obj.kupfer_add_alias(str(other))
            # adding the other's aliases can be misleading
            # since the matched email address might not be
            # what we are e-mailing
            # obj.name_aliases.update(other.name_aliases)
        return obj

class ToplevelGroupingSource (GroupingSource):
    """
    Sources of this type group their leaves with others in the toplevel
    of the catalog.
    """
    _sources = {}

    def __init__(self, name, category):
        GroupingSource.__init__(self, name, [self])
        self.category = category

    def toplevel_source(self):
        if self.category not in self._sources:
            return self
        sources = list(self._sources[self.category].keys())
        return GroupingSource(self.category, sources)

    def initialize(self):
        if not self.category in self._sources:
            self._sources[self.category] = weakref.WeakKeyDictionary()
        self._sources[self.category][self] = 1
        self.output_debug("Register %s source %s" % (self.category, self))

    def finalize(self):
        del self._sources[self.category][self]
        self.output_debug("Unregister %s source %s" % (self.category, self))

class _GroupedItemsSource(Source):
    def __init__(self, leaf):
        Source.__init__(self, str(leaf))
        self._leaf = leaf

    def get_items(self):
        for leaf in self._leaf.links:
            yield leaf

    def repr_key(self):
        return repr(self._leaf)



